from typing import Mapping

from fastapi.params import Param

from apiens.util import decomarker


class fastapi_route(decomarker):
    """ Additional options for FastAPI routes

    Use it to decorate @operation functions and describe HTTP parameters for FastAPI

    Example:
        from apiens import operation
        from apiens.via.fastapi import fastapi_route

        @operation
        @fastapi_route('GET', '/')
        def index():
            pass
    """

    # The HTTP verb to use on this endpoint
    method: str

    # The API path to use
    path: str

    # Customized parameters
    parameters: Mapping[str, Param]

    def __init__(self, /, method: str, path: str, **parameters: Param):
        super().__init__()

        self.method = method
        self.path = path
        self.parameters = parameters


class fastapi_params(fastapi_route):
    """ Additional options for FastAPI class constructor (that is not a route)

    Use it to decorate classes that are not operations themselves,
    but hold parameters common to other operations

    Example:
        @fastapi_params(auth_token=Query(...))
        class UserOperations:
            def __init__(self, auth_token: str = None):
                pass

    """

    MARKER_ATTR = fastapi_route.MARKER_ATTR

    def __init__(self, **parameters: Param):
        # noinspection PyTypeChecker
        super().__init__(None, None, **parameters)





class ObjectMatch:
    """ Unit test helper: an object for == comparisons with other objects field by field.

    Receives custom fields. When compared to another object, performs a field-by-field comparison.
    Only checks the fields from itself; any additional fields that the other object may contain are ignored.

    Example:
         models.User(...) == ObjectMatch(models.User, id=Parameter(), login='kolypto', ...)
    """
    __slots__ = ('_type', '_fields')

    def __init__(self, type, **fields):
        self._type = type
        self._fields = fields

    def __eq__(self, other: object):
        assert isinstance(other, self._type)
        # Iterate our fields only. This will provide the "partial matching" behavior
        for name, value in self._fields.items():
            # will recurse into ObjectMatch.__eq__ if one is encountered
            other_value = getattr(other, name)
            assert value == other_value, f'{other!r} == {self!r} because of {value!r} != {other_value}'
            # if value != getattr(other, name):
            #     return False
        return True

    def __repr__(self):
        return '{type}({fields})'.format(type=self._type.__name__, fields=', '.join(f'{k}={v!r}' for k, v in self._fields.items()))


class DictMatch(dict):
    """ Partial match of a dictionary to another one, with nice diffs

    Example:
        assert DictMatch({'praenomen': 'Gaius'}) == {'praenomen': 'Gaius', 'nomen': 'Julius', 'cognomen': 'Caesar'}
    """
    def __eq__(self, other: dict):
        # Only equal if have the same type
        if not isinstance(other, dict):
            return False

        # Copy the missing values from another dict (in order to have nice diffs)
        self.update({k: v for k, v in other.items() if k not in self})

        # Compare
        return all(other[name] == value for name, value in self.items())


class Parameter:
    """ A parameter that, when compared, grabs the value. Useful for grabbing primary keys. """
    __slots__ = ('_grabbed_value',)

    def __init__(self):
        self._grabbed_value = self.NO_VALUE

    @property
    def value(self):
        assert self._grabbed_value is not self.NO_VALUE, 'NO_VALUE'
        return self._grabbed_value

    def __eq__(self, other):
        # Parameter resolved. Compare. Return result.
        if self._grabbed_value is not self.NO_VALUE:
            assert self._grabbed_value == other
            return self._grabbed_value == other
        # Parameter unresolved. Got a new value. Remember. Compare.
        elif self._grabbed_value is self.NO_VALUE and not isinstance(other, Parameter):
            self._grabbed_value = other
            return True
        # Parameter unresolved. Comparing to another parameter. Ouch.
        else:
            raise AssertionError

    def __repr__(self):
        if self._grabbed_value is self.NO_VALUE:
            return '<NO VALUE>'
        return repr(self._grabbed_value)

    NO_VALUE = object()



from __future__ import annotations
from typing import Mapping, Union, TypedDict, List, Any


class QueryObject(TypedDict):
    """ MongoSQL Query Object """
    project: Union[str, List[str], Mapping[str, int]]
    filter: Mapping[str, Any]
    join: Mapping[str, Mapping]
    joinf: Mapping[str, Mapping]
    sort: Union[str, List[str]]
    group: Union[str, List[str]]
    skip: int
    limit: int
    count: int


class ModernQueryObject(TypedDict):
    """ Modern Query Object """
    select: Union[str, List[Union[str, Mapping]], Mapping[str, Union[int, Mapping]]]
    filter: Mapping[str, Any]
    sort: Union[str, List[str]]
    skip: int
    limit: int



""" Definitions """

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Mapping, Collection, ContextManager, Any, Union, Set, FrozenSet

from .const import InjectionToken, InjectFlags, MISSING


@dataclass
class Resolvable:
    """ A function with dependency information

    The idea is simple: once dependencies are resolved, call `func()` and give it dependencies as `**kwargs`.

    Example:
        Resolvable(
            func=get_user,
            deps_kw={
                'db_connection': Dependency(...),
            }
        )
    """
    __slots__ = 'func', 'deps_kw', 'deps_nopass'

    def __init__(self,
                 func: Callable,
                 deps_kw: Mapping[str, Dependency] = MISSING,
                 deps_nopass: Collection[Dependency] = MISSING):
        self.func = func
        self.deps_kw = deps_kw if deps_kw is not MISSING else {}
        self.deps_nopass = deps_nopass if deps_nopass is not MISSING else []

    # The function to call when the dependencies are resolved
    func: Callable

    # Dependencies to be passed as keyword arguments
    deps_kw: Mapping[str, Dependency]

    # Dependencies to be used without passing as arguments
    deps_nopass: Collection[Dependency]

    def merge(self, another: Resolvable):
        """ Merge another resolvable into this one. This function helps when a function is decorated multiple times. """
        self.deps_kw.update(another.deps_kw)
        self.deps_nopass.extend(another.deps_nopass)
        return self

    def provided_names(self) -> FrozenSet[str]:
        """ Get the names of arguments that will be provided to the function """
        return frozenset(self.deps_kw)


ProviderFunction = Callable[..., Any]
ProviderContextManager = Callable[..., ContextManager[Any]]


@dataclass
class Provider(Resolvable):
    """ A provider for instances that can be obtained from an injector

    Example:
        Provider(
            token=DbSession,
            provider=get_session,
            deps_kw={
                'db_connection': Dependency(...)
            }
        )
    """
    __slots__ = 'token',

    def __init__(self,
                 token: InjectionToken,
                 func: Callable,
                 deps_kw: Mapping[str, Dependency] = MISSING,
                 deps_nopass: Collection[Dependency] = MISSING):
        super().__init__(func=func, deps_kw=deps_kw, deps_nopass=deps_nopass)
        self.token = token

    # Injection token: the dependency it provides
    token: InjectionToken

    # Provider function: either
    # 1. a callable that returns the instance to be provided ; or
    # 2. A context manager, whose __enter__() value is the instance, and __exit__() cleans up when the injector exits
    func: Union[ProviderFunction, ProviderContextManager]

    @classmethod
    def from_resolvable(cls, token: InjectionToken, resolvable: Resolvable):
        """ Provider() from Resolvable() """
        return cls(
            token=token,
            func=resolvable.func,
            deps_kw=resolvable.deps_kw,
            deps_nopass=resolvable.deps_nopass,
        )


@dataclass
class Dependency:
    """ A declaration of depdendency on some value, identified by `token`.

    In order to resolve this dependency, a provider has to be found that provides `token`.

    Example:
        Dependency(
            token=DbSession
        )
    """
    __slots__ = 'token', 'flags', 'default'

    def __init__(self,
                 token: InjectionToken,
                 flags: InjectFlags = InjectFlags.DEFAULT,
                 default: Any = MISSING):
        self.token = token
        self.flags = flags
        self.default = default

    # Injection token: what is required
    token: InjectionToken

    # Injector flags for dependency resolution
    flags: InjectFlags

    # Default value
    default: Any

    @classmethod
    def ensure_dependency(cls, arg: Union[InjectionToken, Dependency]) -> Dependency:
        """ Wrap injection tokens with Dependency() if necessary """
        if isinstance(arg, cls):
            return arg
        else:
            return cls(token=arg)



""" DI errors """

from apiens.di.defs import InjectionToken, Provider


class BaseInjectorError(Exception):
    """ Base for injector errors """


class NoProviderError(BaseInjectorError):
    """ No provider has been found for an injection token

    This error is raised when a dependency is required, but no provider has been found.
    Make sure you've used the correct `token`, and defined a provider for it.
    """

    def __init__(self, message: str, token: InjectionToken):
        super().__init__(message)
        self.token = token


class ClosedInjectorError(BaseInjectorError):
    """ An attempt to operate on an Injector that has already been closed """



""" Tools for patching Python functions' arguments """

import functools
import inspect
import types
from typing import Callable, TypeVar, Any, Generator

ClassmethodT = TypeVar('ClassmethodT', bound=classmethod)
MethodT = TypeVar('MethodT', bound=Callable)


def set_parameter_default(func: Callable, param: str, default: Any) -> Callable:
    """ Set a default value for one function parameter; make all other defaults equal to `...`

    This function is normally used to set a default value for `self` or `cls`:
    weird magic that makes FastAPI treat the argument as a dependency.
    All other arguments become keyword-only, because otherwise, Python won't let this function exist.

    Example:
        set_parameter_default(Cls.method, 'self', Depends(Cls))
    """
    patch = patch_function_keyword_arguments(func)
    try:
        parameter = None
        while parameter := patch.send(parameter):
            # For `name`, replace the default
            if parameter.name == param:
                parameter = parameter.replace(default=default)
    except StopIteration:
        return func


def patch_function_keyword_arguments(func: Callable) -> Generator[inspect.Parameter, inspect.Parameter, Callable]:
    """ A generator for patching a function's parameters.

    NOTE: it makes every argument a keyword-only argument.

    Example:
        patch = patch_function_parameters(endpoint)
        try:
            parameter = None
            while True:
                parameter = parameter.replace(...)
                # Replace this parameter and get the next one
                parameter = patch.send(parameter)
        except StopIteration:
            pass  # Done
    """
    # Get the signature
    sig = inspect.signature(func)

    # Make a new parameter list
    new_parameters = []
    for name, parameter in sig.parameters.items():
        # yield it, and get it back
        parameter = yield parameter

        # make all arguments keyword-only.
        # this makes sure that if a default value has been added to any argument, the Python function will still make sense.
        if parameter.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD):
            # Make them keyword-only
            parameter = parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY)

        # prepare to replace it
        new_parameters.append(parameter)

    # Replace the signature
    setattr(func, '__signature__', sig.replace(parameters=new_parameters))
    return func


def copy_func(f, name=None):
    """ Make a copy of a Python function

    Python copy() does not copy functions. This implementation does it.
    """
    g = types.FunctionType(f.__code__, f.__globals__, name=name or f.__name__, argdefs=f.__defaults__, closure=f.__closure__)
    g = functools.update_wrapper(g, f)  # updates __dict__ as well
    return g



""" DI constants """

from enum import Flag, auto
from typing import NewType, Hashable

# Injection token: the value by which dependency providers are found.
# Normally, a class (associated with its constructor callable)
InjectionToken = NewType('InjectionToken', Hashable)


class InjectFlags(Flag):
    """ Dependency resolution flags """
    # Search for a provider starting from this one, and upwards, towards the root injector
    DEFAULT = auto()

    # Only check the current injector for providers. Fail if not found.
    SELF = auto()

    # Do not check the current injector; start one level higher
    SKIP_SELF = auto()

    # Do not raise errors if no provider is found; return `None`
    OPTIONAL = auto()


# Indication that a value is not provided
MISSING = object()



from .class_based_view import class_based_view, ViewBase
from .class_based_view import api_route, Route
from .class_based_view import fastapi_compatible_method

from .class_based_crud import class_based_crud_view, CrudView



""" A default set of errors for your convenience. Use it if you will. """

from typing import Iterable, List, Union
from http import HTTPStatus


# Init translations
# TODO: will this work with lazy translations?
import gettext
try:
    translation = gettext.translation('apiens')
except FileNotFoundError:
    translation = gettext.NullTranslations()
_ = translation.gettext


from .errors import BaseApplicationError


HTTP_400_BAD_REQUEST = 400
HTTP_401_UNAUTHORIZED = 401
HTTP_403_FORBIDDEN = 403
HTTP_404_NOT_FOUND = 404
HTTP_405_METHOD_NOT_ALLOWED = 405
HTTP_409_CONFLICT = 409
HTTP_500_INTERNAL_SERVER_ERROR = 500
HTTP_501_NOT_IMPLEMENTED = 501


# region HTTP 400 Client Errors

# region 400 Bad Request

class E_BAD_API_REQUEST(BaseApplicationError):
    """ Wrong usage of the API due to technical reasons

    Unlike most other errors, this is a technical error that the user cannot fix.
    """
    httpcode = HTTP_400_BAD_REQUEST
    title = _('Bad API request')
    # Fixit message is common to all errors
    fixit = _('The application has made a wrong query request to the server. '
              'Please contact support and describe the issue.')


class E_API_ARGUMENT(BaseApplicationError):
    """ Wrong argument has been provided

    This is a user input error that the user can understand and fix.

    Args:
        name: The name of the failed argument
    """
    httpcode = HTTP_400_BAD_REQUEST
    title = _('Invalid argument')

    def __init__(self, /, error: str, fixit: str = None, *, name: str, **info):
        """
        Args:
            name: The name of the argument whose value was wrong
        """
        super().__init__(error, fixit, name=name, **info)


class E_API_ACTION(BaseApplicationError):
    """ Wrong action has been requested

    This is a user input error that the user can understand and fix.
    """
    httpcode = HTTP_400_BAD_REQUEST
    title = _('Incorrect action')


class E_CLIENT_VALIDATION(BaseApplicationError):
    """ Validation errors on the provided data

    This is a user input error that the user can understand and fix.

    Info:
        model: The Model the error was found with. Example: 'User'
        errors: The list of errors
    """
    httpcode = HTTP_400_BAD_REQUEST
    title = _('Input validation error')

    def __init__(self, /, error: str = None, fixit: str = None, *, model: str, errors: List[dict], **info):
        """ Iniitialize a validation error

        Args:
            model: Name of the model that has erred
            errors: Errors description: [{ loc: Tuple[str], msg: str, type: str }]
        """
        super().__init__(
            error or _('Invalid input'),
            fixit or _('Please fix the data you have provided and try again'),
            model=model,
            errors=errors,
            **info
        )


    @classmethod
    def from_pydantic_validation_error(cls, pydantic_exception: 'pydantic.ValidationError', error: str = None, fixit: str = None, **info):
        """ Create from Pydantic validation error """
        e = cls(
            error,
            fixit,
            model=pydantic_exception.model.__name__,
            errors=pydantic_exception.errors(),  # [ {loc: Tuple[str], msg: str, type; str} ]
            **info
        )
        return exception_from(e, pydantic_exception)


# region 401 Unauthorized

class E_AUTH_REQUIRED(BaseApplicationError):
    """ The resource requires authentication

    This error means that the user is trying to access a resource without being authenticated.
    They should go to the sign-in place and go through it.
    """
    httpcode = HTTP_401_UNAUTHORIZED
    title = _('Authentication required')


class E_AUTH_CREDENTIALS(BaseApplicationError):
    """ Failed to authenticate the user because of bad input on their side

    This may mean that the login, the password, or whatever else, was invalid.
    """
    httpcode = HTTP_401_UNAUTHORIZED
    title = _('Invalid authentication credentials')


class E_AUTH_USER_DEACTIVATED(BaseApplicationError):
    """ The user account is disabled and cannot be accessed """
    httpcode = HTTP_401_UNAUTHORIZED
    title = _('User account disabled')

# endregion


# region 403 Forbidden

class E_FORBIDDEN(BaseApplicationError):
    """ Trying to perform an action that is forbidden for this user account.

    This error means that the user is signed in, but the action they requested is not allowed for their account.
    """
    httpcode = HTTP_403_FORBIDDEN
    title = _('Action forbidden for this user account')


class E_ROLE_REQUIRED(BaseApplicationError):
    """ Action forbidden because the user does not not have the required role.

    For instance, this error is reported when a non-admin user tries to perform an admin action.
    """
    httpcode = HTTP_403_FORBIDDEN
    title = _('Role required')

    def __init__(self, /, error: str, fixit: str = None, *, required_roles: Iterable[str], **info):
        """
        Args:
            role: The user account type required to access this resource
        """
        super().__init__(error, fixit, required_roles=required_roles, **info)


class E_PERMISSION_REQUIRED(BaseApplicationError):
    """ Action forbidden because the user lacks a permission

    For instance, this error is reported when a user tries to perform an action that requires a special permission to do so.
    """
    httpcode = HTTP_403_FORBIDDEN
    title = _('Permission required')

    def __init__(self, /, error: str, fixit: str = None, *, required_permissions: Iterable[str], **info):
        """
        Args:
            role: The user account type required to access this resource
        """
        super().__init__(error, fixit, required_permissions=required_permissions, **info)

# endregion


# region 404 Not Found

class E_NOT_FOUND(BaseApplicationError):
    """ Object not found

    Info:
        object: The name of the object that was not found
    """
    httpcode = HTTP_404_NOT_FOUND
    title = _('Not found')

    def __init__(self, /, error: str, fixit: str = None, *, object: Union[type, str], **info):
        """

        Args:
            object: The object that has not been found.
        """
        # Convert class names
        if isinstance(object, type):
            object = object.__name__

        # super()
        super().__init__(error, fixit, object=object, **info)


# endregion


# region 409 Conflict

class E_CONFLICT(BaseApplicationError):
    """ Action conflicts with something else """
    httpcode = HTTP_409_CONFLICT
    title = _('Conflict')


class E_CONFLICT_DUPLICATE(E_CONFLICT):
    """ Duplicate id: attempting to create an object with a duplicate field value

    This happens, for instance, when you are trying to create an account with a duplicate phone number
    while the application does not allow that.
    """
    httpcode = HTTP_409_CONFLICT
    title = _('Duplicate entry')


# endregion


# region HTTP 500 Server Errors

class E_FAIL(BaseApplicationError):
    """ Generic server error """
    httpcode = HTTP_500_INTERNAL_SERVER_ERROR
    title = _('Generic server error')


class E_UNEXPECTED_ERROR(BaseApplicationError):
    """ Unexpected error, probably signifying an error in the code or other sort of malfunction """
    httpcode = HTTP_500_INTERNAL_SERVER_ERROR
    title = _('Generic server error')

    @classmethod
    def from_exception(cls, unexpected_exception: BaseException, error: str = None, fixit: str = None, **info):
        """

        Args:
            unexpected_exception: The error that the application did not expect.
        """
        e = cls(
            error or str(unexpected_exception),
            fixit or _('Please try again in a couple of minutes. '
                       'If the error does not go away, contact support and describe the issue'),
            # Adding a `debug` field.
            # It will only be included in non-production mode
            debug_unexpected_exception=str(unexpected_exception),
            **info
        )

        # Link
        return exception_from(e, unexpected_exception)


class E_NOT_IMPLEMENTED(BaseApplicationError):
    """ The method is not yet implemented. """
    httpcode = HTTP_501_NOT_IMPLEMENTED
    title = _('Not implemented')

# endregion


def exception_from(new_exception: BaseException, cause_exception: BaseException):
    """ Link one exception to another to show causal relationship """
    new_exception.with_traceback(cause_exception.__traceback__)
    new_exception.__cause__ = cause_exception
    return new_exception



""" Bindings to the FastAPI framework """
from copy import copy

import fastapi
import fastapi.params
import inspect
from typing import List, Callable, Union, Optional
import urllib.parse

from apiens import operation, di, errors
from apiens.via.fastapi.error_schema import ErrorResponse
from apiens.via.fastapi.fastapi_route import fastapi_route


class OperationalApiRouter(fastapi.APIRouter):
    """ A FastAPI-compatible router to run your @operations

    Example:
        from apiens.via.fastapi import OperationalApiRouter

        router = OperationalApiRouter(debug=True, fully_documented=True)
        router.add_operations(
            index,
            UserOperations,
        )

        # Application
        app = FastAPI()
        app.include_router(router)
    """

    # The injector that's used to run your operations.
    # Don't forget to register your providers on it.
    # Note that this injector is just a template whose providers are used on actual endpoints
    injector: di.Injector

    # Flat operations: ordinary functions
    func_operations: List[Callable]

    # Class-based operations: classes
    class_operations: List[type]

    def __init__(self, *, debug: bool = False, fully_documented: bool = True, **kwargs):
        """

        Args:
            debug: Debug mode; will make extra consistency and quality checks
            fully_documented: Ensure that every function, parameter, and result, has a description and annotation.
                This parameter is dearly loved by documentation freaks, but is probably a curse to everyone else ... :)
            **kwargs: Keyword arguments for the APIRouter, if you for some reason want to tune it to your taste
        """
        super().__init__(**kwargs)

        # Debug mode
        self.debug: bool = debug

        # Do we want all operations fully documented?
        self.fully_documented: bool = fully_documented

        # Init
        self.injector = di.Injector()
        self.func_operations: List[Callable] = []
        self.class_operations: List[type] = []

    def add_operations(self, *operations: Union[Callable, type], prefix: str = None, tags: List[str] = None):
        """ Add operations to the router

        Args:
             *operations: Functions, or classes, decorated by @operation.
             prefix: the prefix to add to every operation
        """
        for operation in operations:
            # Tell classes and functions apart
            if inspect.isclass(operation):
                self.class_operations.append(self.register_class_operations(operation, prefix=prefix, tags=tags))
            else:
                self.func_operations.append(self.register_func_operation(operation, prefix=prefix, tags=tags))

    def register_func_operation(self, func: Callable, *, prefix: str = None, tags: List[str] = None) -> Callable:
        """ Register a single function-operation

        Args:
            func: The function to register as an operation
        """
        # Get the operation
        func_op = operation.get_from(func)
        assert func_op is not None, f'Function {func} must be decorated with @operation'

        # Register
        try:
            self._register_operation(func_op, prefix=prefix, tags=tags)
        except Exception as e:
            raise ValueError(f"Error registering @operation {func_op}") from e

        # Done
        return func

    def register_class_operations(self, class_: type, *, prefix: str = None, tags: List[str] = None) -> type:
        """ Register class-based operations """
        # Get the class operation itself
        class_op = operation.get_from(class_)
        assert class_op is not None, f'Class {class_} must be decorated with @operation'

        # List its sub-operations
        for func_op in operation.all_decorated_from(class_, inherited=True):
            try:
                self._register_operation(func_op, class_op, prefix=prefix, tags=tags)
            except Exception as e:
                raise ValueError(f"Error registering @operation {func_op}") from e

        # Done
        return class_

    def _request_injector(self, func_op: operation) -> di.Injector:
        """ Create a di.Injector() for this request """
        # Init the request Injector by copy()ing the template injector that already has providers set up
        request_injector: di.Injector = copy(self.injector)

        # Provide a few contextual things:
        # * `operation`: the operation that is being run
        # * `di.Injector`: the injector itself, for low-level access
        request_injector.provide_value(operation, func_op)
        request_injector.provide_value(di.Injector, request_injector)

        return request_injector

    def _execute_operation(self, /, func_op: operation, class_op: Optional[operation], operation_kwargs: dict):
        """ Execute an operation, using the di.Injector() from this router

        This function is basically the actual FastAPI endpoint that executes your operations.
        """
        # Within an injector ...
        with self._request_injector(func_op) as request_injector:
            args = []

            # ... init the class
            if class_op:
                instance_kwargs = class_op.pluck_kwargs_from(operation_kwargs)
                instance = request_injector.invoke(class_op.func, **instance_kwargs)

                # Pass the `self` as the first argument
                args.append(instance)

            # ... execute the function
            func_kwargs = func_op.pluck_kwargs_from(operation_kwargs)
            return request_injector.invoke(func_op.func, *args, **func_kwargs)

    def _register_operation(self, func_op: operation, class_op: operation = None, prefix: str = None, tags: List[str] = None):
        # Validate the documentation
        if self.debug:
            func_op.check_operation_documentation(fully_documented=self.fully_documented)

        # Prepare the operation endpoint
        operation_endpoint = self._prepare_function_operation_endpoint(func_op, class_op or None)

        # Choose the method and the path
        method = get_operation_method(func_op)
        path = get_operation_path(func_op)

        if class_op:
            path = path_join(get_operation_path(class_op), path)

        if prefix:
            path = path_join(prefix, path)

        # Choose an operation id
        if class_op:
            operation_id = f'{class_op.operation_id}.{func_op.operation_id}'
        else:
            operation_id = f'{func_op.operation_id}'

        # Register the route
        self.add_api_route(
            # Path: the operation id itself
            path,
            # Func: the function to all
            operation_endpoint,
            # HTTP method. Always 'POST', to let us pass arguments of arbitrary complexity in the body as JSON
            methods=[method],
            # Use the same operation id: openapi-generator will find a good use to it
            operation_id=operation_id,
            name=operation_id,
            tags=tags,
            # Its return type: exactly what the function returns.
            # With some pydantic tuning.
            response_model=func_op.signature.return_type,
            response_model_exclude_unset=True,
            # Documentation.
            **operation_route_documentation_kwargs(func_op)
        )

    def _prepare_function_operation_endpoint(self, func_op: operation, class_op: operation = None) -> Callable:
        """ Make an endpoint-function: to call an operation via a di.Injector()

        This function has to:

        * Receive arbitrary **kwargs to be compatible with any operation's arguments
        * Have a signature that FastAPI can read to infer its arguments and their types
        * Start an injector as a context manager -- to provide dependencies to the operation
        * Run the function
        * Return the result

        In addition, if this function is a method of a class:

        * Create an instance of this class
        * Call the function as a method
        * When calling both, be sure to pluck arguments from **kwargs
        """
        # Prepare the actual endpoint function
        def operation_endpoint(**kwargs):
            return self._execute_operation(func_op, class_op, kwargs)

        # Tune the function's signature
        operations = [class_op, func_op] if class_op else [func_op]
        patch_operation_endpoint_signature(
            operation_endpoint,
            func_op.operation_id,
            func_op.signature.return_type,
            *operations
        )

        # Done
        return operation_endpoint


def patch_operation_endpoint_signature(endpoint: Callable, name: str, return_type: type, *operations: operation) -> Callable:
    """ Prepare an endpoint function that will be understood by FastAPI.

    Goals:

    * Give it a nice name
    * Set the return type annotation
    * Set types and documentation for every parameter

    This method is suitable for patching both a function operation and class operation functions.
    The difference lies in the distinction between function operations and class-based operations:
    * Function operations have arguments which become API input arguments
    * Class-based operations have a constructor that has to be called before the method, and it may also have paramenters.
    Therefore, with class-based operations, we have to merge the parameters from both the constructor and the method.
    """
    # Combine the parameters from all operations
    combined_parameters = {
        argument_name: operation_parameter_info(argument_name, argument_type, op)
        for op in operations
        for argument_name, argument_type in op.signature.arguments.items()
    }

    # Give it a name that might pop up somewhere in tracebacks
    endpoint.__name__ = name

    # Generate a signature
    endpoint.__signature__ = inspect.Signature(
        parameters=list(combined_parameters.values()),
        return_annotation=return_type,
    )

    # Done
    return endpoint


def operation_parameter_info(name: str, type_: type, op: operation) -> inspect.Parameter:
    """ Prepare a parameter for the FastAPI endpoint function """
    # Create the FastAPI parameter
    # In FastAPI, this determines where the parameter is coming from: Body, Query, Path, etc.

    # Is there an override?
    customized = fastapi_route.get_from(op.func)
    if customized and name in customized.parameters:
        parameter = fastapi_route.get_from(op.func).parameters[name]
    # Use the default: Body()
    else:
        parameter = fastapi.Body(
            # If the parameter has a default, use it.
            # If not, use the `...`. This is how FastAPI declares required parameters
            op.signature.argument_defaults.get(name, ...),
            # Documentation for the parameter
            title=op.doc.parameters[name].summary if name in op.doc.parameters else None,
            description=op.doc.parameters[name].description if name in op.doc.parameters else None,
        )

    # For `Body` parameters, make sure they have `embed=True`.
    # That is, don't "unwrap" its properties into the top level of the body, but keep it as a named field
    if isinstance(parameter, fastapi.params.Body):
        parameter.embed = True

    # Make the parameter
    return inspect.Parameter(
        # Argument name: something that the API user has to provide
        name=name,
        # All arguments are keyword arguments
        kind=inspect.Parameter.KEYWORD_ONLY,
        # Parameter type annotation
        annotation=type_,
        # This is how FastAPI declares dependencies: <arg name> = Body() / Query() / Path()
        default=parameter
    )


def operation_route_documentation_kwargs(op: operation) -> dict:
    """ Get documentation kwargs for the route

    This function reads the docstring from `func` and returns kwargs for `add_api_route()`
    """
    route_kw = {}

    # Function documentation
    if op.doc.function:
        route_kw['summary'] = op.doc.function.summary
        route_kw['description'] = op.doc.function.description

    if op.doc.deprecated:
        route_kw['deprecated'] = True
        # TODO: include more info? `op.doc.deprecated.version/summary/description` & stuff

    # Result documentation
    if op.doc.result:
        route_kw['response_description'] = (
            # Got to put them together because we have 2 fields, but FastAPI has only one
            (op.doc.result.summary or '') +
            (op.doc.result.description or '')
        )

    # Errors documentation.
    # Unfortunately, OpenAPI's responses are only described using http codes.
    # We will have to group our errors by HTTP codes :(
    if op.doc.errors:
        route_kw['responses'] = responses = {}

        # For every error
        for error_type, error_doc in op.doc.errors.items():
            # Only handle the types we know
            if not issubclass(error_type, errors.BaseApplicationError):
                continue

            # Convert
            httpcode = error_type.httpcode

            # Init the HTTP code object
            if httpcode not in responses:
                responses[httpcode] = {
                    'model': ErrorResponse,
                    'description': '',
                }

            # Add our error to this description as some sort of Markdown list.
            # That's probably the best we can do here.
            responses[httpcode]['description'] += (
                f'`{error_type.__name__}`. {error_doc.summary}{error_doc.description or ""}\n\n'
            )

    # Done
    return route_kw


def get_operation_method(op: operation) -> str:
    """ Having an @operation, decide which HTTP method to use for it. It may be overridden with @fastapi_route() """
    if fastapi_route.is_decorated(op.func):
        return fastapi_route.get_from(op.func).method
    else:
        return 'POST'


def get_operation_path(op: operation) -> str:
    """ Having an @operation, decide which path to use for it. It may be overridden with @fastapi_route() """
    if fastapi_route.is_decorated(op.func):
        return path_join('/', fastapi_route.get_from(op.func).path)
    else:
        return path_join('/', op.operation_id)


def path_join(base: str, url: str) -> str:
    """ Join two parts of an URL """
    base = base.rstrip('/') + '/'
    url = url.lstrip('/')
    return urllib.parse.urljoin(base, url)



from copy import deepcopy
from typing import TypeVar

from sqlalchemy.orm.attributes import InstrumentedAttribute, ScalarObjectAttributeImpl
from sqlalchemy.orm.base import instance_state, DEFAULT_STATE_ATTR
from sqlalchemy.orm.state import InstanceState

from sa2schema import sa_model_info, AttributeType
from sa2schema.info import RelationshipInfo, PropertyInfo, HybridPropertyInfo


SAInstanceT = TypeVar('SAInstanceT')


class InstanceHistoryProxy:
    """ Proxy object to access historical model attributes' values.

    This lets the API save hook compare values of the modified instance to the values of an unmodified one.

    Example:
        # Load, back-up
        instance = ssn.query(...).one()
        prev = InstanceHistoryProxy(instance)

        # Modify
        instance.field = 'value'

        # ... somewhere inside a save hook ...
        # Compare
        assert prev.field != instance.field
    """
    def __init__(self, instance: object, *, copy: bool = False):
        """ Make a lightweight snapshot of an instance.

        Be sure to do it before flush(), because flush() will erase all in-memory changes.

        Args:
            instance: The instance to get the historical values for.
            copy: Copy every mutable value.
                Useful for embedded dictionaries, but it a bit more expensive, so disabled by default.
        """
        # Model info
        self.__model_info = sa_model_info(type(instance), types=AttributeType.ALL)

        # Remember the historical values
        self.__state: InstanceState = instance_state(instance)
        self.__historical = {
            # Merging dictionaries is very cheap
            **self.__state.dict,  # current values
            **self.__state.committed_state,  # overwritten with DB values
        }

        # Make a deep copy to preserve embedded dictionaries
        if copy:
            self.__historical = deepcopy(self.__historical)

    def __getattr__(self, attr_name):
        # _sa_instance_state() may be requested when accessing relatipnships (because they are descriptors)
        if attr_name == DEFAULT_STATE_ATTR:
            # Initialize it lazily and only once
            return self.__enable_relationships_access()

        # Attribute info
        try:
            attr_info = self.__model_info[attr_name]
        except KeyError as e:
            raise AttributeError(attr_name) from e

        # Relationships
        if isinstance(attr_info, RelationshipInfo):
            # Relationships are descriptors, and so they need to be called properly
            attr: InstrumentedAttribute = getattr(self.__state.class_, attr_name)
            #return attr.__get__(self, self.__state.class_)  # cannot do it because it tries to use the self.__dict__ which is empty
            rel: ScalarObjectAttributeImpl = attr.impl
            return rel.get_committed_value(self.__state, self.__historical)
        # @property
        elif isinstance(attr_info, (PropertyInfo, HybridPropertyInfo)):
            # Because properties may use other columns,
            # we have to run it against our`self`, because only then it'll be able to get the original values.
            prop: property = getattr(self.__state.class_, attr_name)
            return prop.fget(self)
        # Loaded columns
        elif attr_name in self.__historical:
            return self.__historical[attr_name]
        # Unloaded columns
        else:
            # Get directly from the instance.
            # May trigger lazy loads. If this is undesirable, you should have used raiseload()
            return getattr(self.__state.object, attr_name)

    def __enable_relationships_access(self) -> InstanceState:
        # Install the internal SqlAlchemy's property on our ModelHistoryProxy to mimic the original object.
        # This is necessary for relationship attributes to work, because they are descriptors
        # Thanks @vihtinsky for this magic
        new_state = InstanceState(self, self.__state.manager)
        new_state.key = self.__state.key
        new_state.session_id = self.__state.session_id
        setattr(self, DEFAULT_STATE_ATTR, new_state)
        return new_state


def get_history_proxy_for_instance(instance: SAInstanceT, copy: bool = False) -> SAInstanceT:
    """ Get a permanent InstanceHistoryProxy for an instance.

    Every time this function is called on an instance, even after flush, the very same InstanceHistoryProxy will be returned.
    Be careful with long-living instances: they will remember their original values the whole time.
    """
    state: InstanceState = instance_state(instance)

    # Create a new one
    if InstanceHistoryProxy not in state.info:
        state.info[InstanceHistoryProxy] = InstanceHistoryProxy(instance, copy=copy)

    # Done
    return state.info[InstanceHistoryProxy]



""" Schemas for the Error response """
import pydantic as pd
from typing import Optional

from apiens.errors import BaseApplicationError


class ErrorObject(pd.BaseModel):
    """ Object representing an API error """
    # Generic fields
    name: str = pd.Field(...,
                         title='Error class name.',
                         description='Use it in your code to handle different error conditions.')

    title: str = pd.Field(...,
                          title='A generic descriptive message',
                          description='Not for the user, but for the developer')
    httpcode: int = pd.Field(...,
                             title='HTTP error code')

    # Specific error data
    error: str = pd.Field(...,
                          title='Error message for the user',
                          description='Negative side: what has gone wrong')
    fixit: str = pd.Field(...,
                          title='Suggested action for the user',
                          description='Positive side: what the user is supposed to do')
    info: dict = pd.Field(...,
                          title='Additional information',
                          description='Structured information about the error')

    # Debug information
    debug: Optional[dict] = pd.Field(...,
                                     title='Additional debug information',
                                     description='Only available in non-production mode')

    @classmethod
    def from_exception(cls, e: BaseApplicationError, include_debug_info: bool):
        """ Create the `ErrorObject` from a Python exception """
        return cls(
            name=e.name,
            title=str(e.title),
            httpcode=e.httpcode,
            error=str(e.error),
            fixit=str(e.fixit),
            info=e.info,
            debug=None if include_debug_info else e.debug,
        )


class ErrorResponse(pd.BaseModel):
    """ Error response as returned by the API """
    error: ErrorObject

    @classmethod
    def from_exception(cls, e: BaseApplicationError, include_debug_info: bool):
        """ Create the `ErrorResponse` from a Python exception """
        return ErrorResponse(error=ErrorObject.from_exception(e, include_debug_info=include_debug_info))



from functools import lru_cache
from inspect import unwrap, isfunction
from typing import ClassVar, Callable, TypeVar, Optional, Sequence, Any, Type, Union

Cls_T = TypeVar('Cls_T')
Self_T = TypeVar('Self_T')


class decomarker_meta(type):
    """ Metaclass that supports simple isinstance() checks for func decorators """

    def __instancecheck__(self: 'decomarker', instance: Any):
        return isfunction(instance) and self._has_marker(instance)


class decomarker(metaclass=decomarker_meta):
    """ A transparent decorator that adds metainformation to a function.

    Features:
    * Add information to a callable
    * instanceof() check lets you check whether a function has been decorated or not
    * Completely transparent: a decorator that does not wrap
    * When used on class methods, allows you to collect all marked methods from that class

    Use it to implement custom behaviors on top of other methods.

    Example:

        class known_errors(decomarker):
            known_errors: dict

            def __init__(self, known_errors: dict):
                super().__init__()
                self.known_errors = known_errors

        @known_errors({
            Exception: 'just happens from time to time',
        })
        def func():
            raise Exception('Oops')

        func_errors = known_errors.get_from(func)

    You can even decorate classes:

        @known_errors({ ... })
        class UserVew:
            ...

    But in this case, you have to consistently decide whether you decorate the class itself, or its constructor.
    """

    # Name of the custom attribute installed on decorated methods.
    # This marker is what makes them stand out.
    # You don't have to set it; it's automatically named after the class
    MARKER_ATTR: ClassVar[str]

    # Decorated func
    func: Callable = None
    func_name: str = None

    def __init__(self):  # override me to receive arguments
        pass

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # Initialize the marker if not already:
        if getattr(cls, 'MARKER_ATTR', None) is None:
            cls.MARKER_ATTR = '::' + cls.__name__

    def __call__(self, func):
        return self.decorator(func)

    def decorator(self, func: Union[Callable, type]):
        # Make sure this decorator is callable only once
        assert self.func is None

        # Function decorated multipl times? Merge.
        prev_marker = self.get_from(func)
        if prev_marker is not None:
            self._merge(prev_marker)

        # Mark the func. Use unwrap() to make sure we get to the meat.
        setattr(unwrap(func), self.MARKER_ATTR, self)

        # Remember & Return
        self.func = func
        self.func_name = func.__name__
        return func

    def _merge(self: Self_T, another: Self_T):
        cls_name = self.__class__.__name__
        raise NotImplementedError(f'@{cls_name} does not support re-decoration')

    def __repr__(self):
        return f'@{self.__class__.__name__}({self.__dict__!r})'

    @classmethod
    def _has_marker(cls, func: Callable) -> bool:
        """ Is the given function decorated with this decorator """
        return hasattr(unwrap(func), cls.MARKER_ATTR)

    # region Public API

    @classmethod
    def is_decorated(cls, func: Callable) -> bool:
        """ Check whether the given func is decorated with @cls()

        Works with re-decorated methods, but only if update_wrapper() was properly used
        """
        return isinstance(func, cls)

    @classmethod
    def get_from(cls: Type[Cls_T], func: Callable) -> Optional[Cls_T]:
        """ Get the @cls decorator object of the wrapped func """
        return getattr(unwrap(func), cls.MARKER_ATTR, None)

    @classmethod
    @lru_cache()
    def all_decorated_from(cls: Type[Cls_T], class_: type, *, inherited: bool = False) -> Sequence[Cls_T]:
        """ Get all decorated methods from a class.

        This method can be used when decorated methods provide some sort of group behavior.

        Args:
            class_: the class to collect decorated methods from
            inherited: whether to include methods decorated on parent classes
        """
        assert isinstance(unwrap(class_), type)

        if inherited:
            members = [getattr(class_, attr) for attr in dir(class_)]
        else:
            members = class_.__dict__.values()

        return tuple(cls.get_from(member)
                     for member in members
                     if isinstance(member, cls) and type(member) != cls)

    # endregion



import mongosql
import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.orm.attributes
from typing import Type, Optional, ClassVar, Union, Tuple

from apiens.views.crud.crud_settings import CrudSettings
from apiens.views.mongoquery_crud.defs import QueryObject


class MongoCrudSettings(CrudSettings):
    """ Settings for MongoSql-enabled Crud Base

    Expectations:
    * crud.query_object: MongoSql query object
    * (optional) crud._mongoquery(_mongoquery: MongoQuery): a hook to customize the MongoQuery before it's finalized
    """
    # MongoQuery class to use for queries. Possible to customize.
    MONGOQUERY_CLASS: ClassVar[Type[mongosql.MongoQuery]] = mongosql.MongoQuery

    # MongoSQL default query object
    _query_defaults: Optional[dict] = None

    # Load only for CRUD operations
    _load_only: Tuple[Union[str, sa.orm.attributes.InstrumentedAttribute]] = ()

    # The MongoQuery object that's going to be used for new queries
    _reusable_mongoquery: Optional[mongosql.MongoQuery] = None

    # Configure

    def load_only(self, *fields: Union[str, sa.orm.attributes.InstrumentedAttribute]):
        """ Configure the names of instance fields that will be loaded for your CRUD handler.

        Only name fields that you're actually going to use in your update() and delete() operations

        Args:
            *fields: Names of fields, attributes, or columns
        """
        self._load_only = fields
        return self

    def query_defaults(self, **query_defaults):
        """ Configure MongoSQL auerying

        Args:
            query_defaults: MongoSQL query defaults
        """
        self._query_defaults = query_defaults
        return self

    def mongosql_config(self, **mongoquery_settings):
        """ Configure MongoSQL auerying

        Args:
            **mongoquery_settings: Settings for MongoQuery.
                See `MongoQuerySettingsDict`.
        """
        self._reusable_mongoquery = mongosql.Reusable(  # noqa
            self.MONGOQUERY_CLASS(self.Model, mongoquery_settings)
        )
        return self

    # Is configured?

    @property
    def _mongoquery_enabled(self):
        """ Check: MongoQuery configured and enabled? """
        return self._reusable_mongoquery is not None

    # Query

    def _mongoquery(self, from_query: sa.orm.Query, query_object: Optional[QueryObject] = None):
        """ Start a MongoSql query using the provided QueryObject and Query

        Args:
            from_query: SqlAlchemy Query to start a MongoSql query from
            query_object: MongoSql Query Object
        """
        # Merge the defaults in
        query_obj = {
            **(self._query_defaults or {}),
            **(query_object or {})
        }

        # Go
        return self._reusable_mongoquery.from_query(from_query).query(**query_obj)



import psycopg2
import psycopg2.errors
import psycopg2.extensions
from functools import lru_cache
from sqlalchemy import exc as sa_exc
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.sql.schema import UniqueConstraint, Table, Column, Index
from typing import Optional, Union, Tuple


# TODO: handle other errors?
# Class 23 — Integrity Constraint Violation
# 23000	    integrity_constraint_violation  IntegrityConstraintViolation
# 23001	    restrict_violation              RestrictViolation
# 23502	    not_null_violation              NotNullViolation
# 23503	    foreign_key_violation           ForeignKeyViolation
# 23505	    unique_violation                UniqueViolation
# 23514	    check_violation                 CheckViolation
# 23P01	    exclusion_violation             ExclusionViolation


def extract_postgres_unique_violation_columns(err: sa_exc.IntegrityError, Base: DeclarativeMeta) -> Tuple[Column]:
    """ Having an IntegrityError with a unique index violation, extract the list of offending columns

    Args:
        err: The IntegrityError exception

    Returns:
        The list of columns involved in a unique index violation. Typically just one.

    Raises:
        ValueError: not a unique index violation
    """
    assert isinstance(err.orig, psycopg2.errors.UniqueViolation)

    # Get diagnostic info
    diag: psycopg2.extensions.Diagnostics = err.orig.diag

    # Find the schema item
    # Whatever this `schema_item` object is, it has a `.columns` property
    table: Table = Base.metadata.tables[diag.table_name]
    schema_item = get_unique_constraint_or_index_by_name(table, diag.constraint_name)

    # List those columns
    return tuple(schema_item.columns)


# Whatever this object is, it has the `.columns` property
SchemaItemWithColumnsProperty = Union[UniqueConstraint, Index]  # SchemaItem, ColumnCollectionMixin


@lru_cache()
def get_unique_constraint_or_index_by_name(table: Table, name: str) -> SchemaItemWithColumnsProperty:
    """ Find a UniqueConstraint or a unique Index by name

    When an IntegrityError is raised, it does not report a column name; it reports a constraint name, like this:

        (psycopg2.errors.UniqueViolation) duplicate key value violates unique constraint "uq__users__login"

    The postgres error object, however, contains the constraint name:

        diag: psycopg2.extensions.Diagnostics = err.orig.diag
        diag.constraint_name

    This function takes this name, and tries to find a corresponding SQLALchemy schema element,
    either UniqueConstraint or Index.

    Usage:

        schema_item = get_unique_constraint_or_index_by_name(models.User.__table__, 'uq__users__login')
        schema_item.columns

    Raises:
        KeyError: no constraint or index is found
    """
    # Try to find a constraint
    constraint = _get_unique_constraint_by_name(table, name)
    if constraint:
        return constraint

    # Try to find a unique index
    index = _get_unique_index_by_name(table, name)
    if index:
        return index

    # Not found
    raise KeyError(name)


def _get_unique_constraint_by_name(table: Table, constraint_name: str) -> Optional[UniqueConstraint]:
    for constraint in table.constraints:
        if isinstance(constraint, UniqueConstraint) and constraint.name == constraint_name:
            return constraint
    return None


def _get_unique_index_by_name(table: Table, index_name: str) -> Optional[Index]:
    for index in table.indexes:
        if index.unique and index.name == index_name:
            return index
    return None



""" FastAPI bindings to CrudBase crud handlers

Example:
    class Response(pd.BaseModel):
        user: UserCrud.crudsettings.ResponseSchema

    class ListResponse(pd.BaseModel):
        users: List[UserCrud.crudsettings.ResponseSchema]

    @class_based_crud_view(router)
    class UserView(CrudView):
        CrudHandler = UserCrud
        name = 'user'
        names = 'users'

        Response: ClassVar[Type[pd.BaseModel]] = Response
        ListResponse: ClassVar[Type[pd.BaseModel]] = ListResponse

        def __init__(self, ssn: Session = Depends(db_session)):
            self.crud = self.CrudHandler(ssn)

        def primary_key(self, id: int = Path(..., title='User id')) -> dict:
            return {'id': id}

        create_or_update = NotImplemented

Is approximately equivalent to the following code:

    @router.get('/', response_model=ListResponse, response_model_exclude_unset=True)
    def list_users(ssn: Session = Depends(db_session)):
        with UserCrud(ssn).transaction() as crud:
            users = crud.list()
            return ListResponse(users=users)

    @router.get('/{id}', response_model=Response, response_model_exclude_unset=True)
    def get_user(id: int, ssn: Session = Depends(db_session)):
        with UserCrud(ssn).transaction() as crud:
            user = crud.get(id=id)
            return Response(user=user)

    @router.post('/', response_model=Response, response_model_exclude_unset=True)
    def create_user(user: UserCrud.crudsettings.CreateInputSchema = Body(..., embed=True),
                    ssn: Session = Depends(db_session)):
        with UserCrud(ssn).transaction() as crud:
            user = crud.create(user)
            return Response(user=user)

    @router.post('/{id}', response_model=Response, response_model_exclude_unset=True)
    def update_user(id: int,
                    user: UserCrud.crudsettings.UpdateInputSchema = Body(..., embed=True, title='Partial User: only modified fields'),
                    ssn: Session = Depends(db_session)):
        with UserCrud(ssn).transaction() as crud:
            user = crud.update(user, id=id)
            return Response(user=user)

    @router.delete('/{id}', response_model=Response, response_model_exclude_unset=True)
    def delete_user(id: int, ssn: Session = Depends(db_session)):
        with UserCrud(ssn).transaction() as crud:
            user = crud.delete(id=id)
            return Response(user=user)
"""

from copy import copy

import pydantic as pd
from apiens.views.mongoquery_crud import MongoCrudBase, MongoCrudSettings
from fastapi import APIRouter
from fastapi import Body, Depends
from typing import Type, ClassVar, Callable, Mapping, List, Optional

from .class_based_view import ViewBase  # noqa
from .class_based_view import class_based_view, Route, patch_method_fastapi_compatible
from .patch_func import patch_function_keyword_arguments


class CrudView:
    """ A class-based view for CrudBase crud handlers.

    An instance is created on every request, and one of its methods is executed.
    """

    # The Crud Handler to use
    CrudHandler: ClassVar[Type[MongoCrudBase]]

    # How to call the field in the response?
    # `name` when a single object is returned;
    # `names` when multiple objects are returned
    name: ClassVar[str]
    names: ClassVar[str]

    # Response schemas:
    # `Response` when a single object is returned
    # `ListResponse` when multiple objects are returned
    # Schemas will be generated using create_model() and `name` with `names` if not provided
    Response: ClassVar[Type[pd.BaseModel]]
    ListResponse: ClassVar[Type[pd.BaseModel]]

    # `crud`: instantiated CrudHandler
    __slots__ = 'crud',
    crud: MongoCrudBase

    def __init__(self):
        """ View constructor, executed on every request. Accepts dependencies as arguments. """
        # Override this method to declare your real dependencies
        raise NotImplementedError

    def primary_key(self) -> dict:
        """ Get request arguments and convert them to crud kwargs

        The output of this method is given to update() and delete() crud operations as kwargs.
        """
        # Override this method: declare the primary key dependencies and return them as a dict
        raise NotImplementedError

    # region CRUD methods

    # TODO: Currently, FastAPI uses `Response` and `ListResponse` to actually validate the resulting objects.
    #  Make this behavior optional: do not validate responses in production!

    # NOTE: return annotation is a forward reference that is dynamically re-evaluated on a per-class basis.
    def list(self) -> 'ListResponse':
        """ list() CRUD operation """
        with self.crud.transaction() as crud:
            items = crud.list()
            return self.ListResponse(**{self.names: items})

    # NOTE: Depends(primary_key) will be replaced by your actual primary key function
    def get(self, primary_key: dict = Depends(primary_key)) -> 'Response':
        """ get() CRUD operation """
        with self.crud.transaction() as crud:
            item = crud.get(**primary_key)
            return self.Response(**{self.name: item})

    # NOTE: `item` model and `alias` will be replaced from your `name` and `CrudHandler`
    def create(self, item: pd.BaseModel = Body(..., alias=None, embed=True, title='The new object to create')) -> 'Response':
        """ create() CRUD operation """
        with self.crud.transaction() as crud:
            item = crud.create(item)
            return self.Response(**{self.name: item})

    # NOTE: Depends(primary_key) will be replaced by your actual primary key function
    # NOTE: `item` model and `alias` will be replaced from your `name` and `CrudHandler`
    def update(self,
               primary_key: dict = Depends(primary_key),
               item: pd.BaseModel = Body(..., alias=None, embed=True, title='Partial object: only modified fields')) -> 'Response':
        """ update() CRUD operation """
        with self.crud.transaction() as crud:
            item = crud.update(item, **primary_key)
            return self.Response(**{self.name: item})

    # NOTE: Depends(primary_key) will be replaced by your actual primary key function
    def delete(self, primary_key: dict = Depends(primary_key)) -> 'Response':
        """ delete() CRUD operation """
        with self.crud.transaction() as crud:
            item = crud.delete(**primary_key)
            return self.Response(**{self.name: item})

    # NOTE: `item` model and `alias` will be replaced from your `name` and `CrudHandler`
    def create_or_update(self, item: pd.BaseModel = Body(..., alias=None, embed=True,
                                                         title='The object to save',
                                                         description='It will be updated if found by the primary key; created otherwise')) -> 'Response':
        """ create_or_update() CRUD operation """
        with self.crud.transaction() as crud:
            item = crud.create_or_update(item)
            return self.Response(**{self.name: item})

    # endregion


class class_based_crud_view(class_based_view):
    """ Decorator for the class-based CrudView """

    def __init__(self, router: APIRouter, pk_path: str = '/{id:int}', *routes: Route, operation_id_fmt: Optional[str] = '{method}_{name}'):
        """ Add routes from this decorated class to `router`

        Args:
            router: The APIRouter to register the routes on
            pk_path: URL path to use for a route with a primary key. Example: '/{id}'
            *routes: Additional routes to register
            operation_id_fmt: OperationId format
        """
        root_path = '/'
        super().__init__(
            router,
            # Known API routes.
            # If some of them are `None` or `NotImplemented`, they will be ignored
            Route('list', root_path, 'GET'),
            Route('get', pk_path, 'GET'),
            Route('create', root_path, 'POST'),
            Route('create_or_update', root_path, 'PUT'),
            # NOTE: ReST standards use PUT for update(). I prefer POST.
            Route('update', pk_path, 'POST'),
            Route('delete', pk_path, 'DELETE'),
            # Extra routes
            *routes
        )
        self.operation_id_fmt = operation_id_fmt

    def _prepare_class(self, ViewCls: Type[CrudView]):
        # Generate default `Response` and `ListResponse`
        if not getattr(ViewCls, 'Response', None):
            ViewCls.Response = self._prepare_class_generate_Response(ViewCls)
        if not getattr(ViewCls, 'ListResponse', None):
            ViewCls.ListResponse = self._prepare_class_generate_ListResponse(ViewCls)

        # super
        super()._prepare_class(ViewCls)

    def _patch_route_endpoints(self, ViewCls: Type[CrudView]):
        super()._patch_route_endpoints(ViewCls)

        # Prepare routes as a mapping
        routes_map = {route.view_name: route for route in self.routes}
        # Patch routes for compatibility with FastAPI
        self._patch_route_endpoints_for_fastapi(ViewCls, routes_map)

    def _patch_route_endpoints_for_fastapi(self, ViewCls: Type[CrudView], routes_map: Mapping[str, Route]):
        # Patch the `primary_key()` method: declare `self` to be a dependency to make sure it works
        ViewCls.primary_key = patch_method_fastapi_compatible(ViewCls, copy(ViewCls.primary_key))
        
        # This is how we generate operation_id
        opid = lambda method, name: (
            self.operation_id_fmt.format(method=method, name=name)
            if self.operation_id_fmt else
            None
        )

        # Patch every view to make sure it works as a CRUD handler
        # Each endpoint can receive: primary_key(), `item` input item with `InputSchema` annotation
        crudsettings: MongoCrudSettings = ViewCls.CrudHandler.crudsettings
        self._patch_route_endpoint(routes_map, 'list', opid('list', ViewCls.names))
        self._patch_route_endpoint(routes_map, 'get', opid('get', ViewCls.name))
        self._patch_route_endpoint(routes_map, 'create', opid('create', ViewCls.name),
                                   InputSchema=crudsettings.CreateInputSchema)
        self._patch_route_endpoint(routes_map, 'update', opid('update', ViewCls.name),
                                   InputSchema=crudsettings.UpdateInputSchema)
        self._patch_route_endpoint(routes_map, 'create_or_update', opid('create_or_update', ViewCls.name),
                                   InputSchema=crudsettings.CreateOrUpdateInputSchema)
        self._patch_route_endpoint(routes_map, 'delete', opid('delete', ViewCls.name))

    @classmethod
    def _patch_route_endpoint(cls,
                              routes_map: Mapping[str, Route],
                              view_name: str,
                              operation_id: Optional[str], *,
                              InputSchema: Type[pd.BaseModel] = None):
        """ Patch a route's endpoint to make it compatible with FastAPI.

        The following patches are done to views that require it:

        * A view requires the `primary_key`: Depends(ViewCls.primary_key) is provided
        * A view requires the `item` input data: Body(alias=ViewCls.name) is provided
        """
        # Get the route. Skip it if not implemented.
        route = routes_map[view_name]
        if not route.view_func_implemented:
            return

        # Patch it
        ViewCls: Type[CrudView] = route.ViewCls  # noqa
        route.view_func = patch_crud_handler_endpoint_arguments(
            endpoint=route.view_func,
            # Patches
            primary_key=ViewCls.primary_key,  # primary_key() function
            input_item_name=ViewCls.name,
            InputSchema=InputSchema,
        )

        # Give it a nice name (if not already set)
        if operation_id:
            route.api_route_kwargs.setdefault('operation_id', operation_id)

        # Additional route settings
        route.api_route_kwargs.update(
            # Not setting `response_model` because it comes from return annotations
            #response_model=ListResponse or Response
            response_model_exclude_unset=True,
        )

    def _prepare_class_generate_Response(self, ViewCls: Type[CrudView]) -> Type[pd.BaseModel]:
        """ Generate the ViewCls.Response schema (when not provided) """
        return pd.create_model(
            # module/view/class combination has to be unique.
            # this enables us to have multiple views in a file
            ViewCls.__name__ + 'Response',
            __module__=ViewCls.__module__,
            **{ViewCls.name: (
                ViewCls.CrudHandler.crudsettings.ResponseSchema,
                pd.Field(...))}
        )

    def _prepare_class_generate_ListResponse(self, ViewCls: Type[CrudView]) -> Type[pd.BaseModel]:
        """ Generate the ViewCls.ListResponse schema (when not provided) """
        return pd.create_model(
            ViewCls.__name__ + 'ListResponse',
            __module__=ViewCls.__module__,
            **{ViewCls.names: (
                List[ViewCls.CrudHandler.crudsettings.ResponseSchema],
                pd.Field(...))}
        )


def patch_crud_handler_endpoint_arguments(
        endpoint: Callable, *,
        primary_key: Callable = None,
        input_item_name: str = None,
        InputSchema: Type[pd.BaseModel] = None):
    """ Take an endpoint and patch its arguments:

    * `pk`: make it use the cls.primary_key
    * `item`: give it a proper `cls.name` and a type annotation
    
    Args:
        endpoint: The endpoint: view function to patch
        primary_key: Dependency provider for the `pk` dependency: a function
        input_item_name: Field name for the input object (create() and update())
        InputSchema: Pydantic schema for the input object
    """
    # Get a patched copy
    new_endpoint = copy(endpoint)
    patch = patch_function_keyword_arguments(new_endpoint)
    try:
        parameter = None  # to make sure that the first send() gets a None
        while parameter := patch.send(parameter):
            # Primary key
            # Replace: set default value = Depends(primary_key)
            # This is necessary because the `primary_key` function is local to every implementation
            if parameter.name == 'primary_key':
                assert primary_key
                parameter = parameter.replace(default=Depends(primary_key))
            # Input item
            # Replace: set type annotation = InputSchema and default value's alias to `name`
            # This input item's name and schema depend on the implementation
            elif parameter.name == 'item':
                assert InputSchema, f'No input schema provided for {endpoint} in the relevant Crud Handler settings. Cannot configure.'
                assert input_item_name, f'No inpput object `name` was provided. Cannot configure {endpoint}'
                default = copy(parameter.default)
                default.alias = input_item_name
                parameter = parameter.replace(default=default, annotation=InputSchema)
    except StopIteration:
        # Done: return the patched endpoint
        return new_endpoint



import pydantic as pd
import sqlalchemy as sa
from functools import cached_property
from sqlalchemy.ext.declarative import DeclarativeMeta
from typing import Type, Optional, Tuple, Iterable, FrozenSet, Set, Mapping

import sa2schema
from . import crudbase
from .defs import AUTOMATIC

PydanticModelT = Type[pd.BaseModel]


class CrudSettings:
    """ Crud Settings object """
    # Debug mode?
    _debug: bool = False

    # The SqlAlchemy model this settings object is made for
    Model: DeclarativeMeta

    # Primary key columns for the model.
    _primary_key: Tuple[str] = AUTOMATIC

    # Is the class using natural primary keys?
    _natural_primary_key: bool = False

    # Custom excluded fields
    _extra_exclude_on_create: FrozenSet[str] = frozenset()
    _extra_exclude_on_update: FrozenSet[str] = frozenset()

    # Detailed model fields info
    _field_names_configured: bool = False
    _ro_fields: Optional[FrozenSet[str]] = None
    _rw_fields: Optional[FrozenSet[str]] = None
    _const_fields: Optional[FrozenSet[str]] = None

    def __init__(self,
                 Model: DeclarativeMeta, *,
                 ResponseSchema: PydanticModelT,
                 CreateSchema: Optional[PydanticModelT],
                 UpdateSchema: Optional[PydanticModelT],
                 debug: bool = False
                 ):
        """
        Args:
            Model: the SqlAlchemy model to work with
            ResponseSchema: Pydantic model for returned objects
            CreateSchema: Pydantic input model for creating objects
            UpdateSchema: Pydantic input model for updating objects
            debug: Debug mode (for testing and non-production instances)
        """
        self._debug = debug

        # SqlAlchemy
        self.Model = Model

        # Pydantic schemas: input
        self.CreateInputSchema: Optional[PydanticModelT] = CreateSchema
        self.UpdateInputSchema: Optional[PydanticModelT] = UpdateSchema
        self.CreateOrUpdateInputSchema: Optional[PydanticModelT] = None

        # Pydantic schemas: output
        # NOTE: these schemas are all the same. If you need to customized them, use customize()
        self.ResponseSchema: PydanticModelT = ResponseSchema  # same for every response, unless overridden
        self.GetResponseSchema: PydanticModelT = ResponseSchema
        self.ListResponseSchema: PydanticModelT = ResponseSchema
        self.CreateResponseSchema: PydanticModelT = ResponseSchema
        self.UpdateResponseSchema: PydanticModelT = ResponseSchema
        self.DeleteResponseSchema: PydanticModelT = ResponseSchema
        self.CreateOrUpdateResponseSchema: Optional[PydanticModelT] = None

        # Primary key
        self._primary_key = sa2schema.sa_model_primary_key_names(self.Model)

    # region Settings

    # Methods to further refine the settings

    def debug(self, debug: bool):
        """ Enable debug mode?"

        In debug mode, Crud handlers may make additional checks and throw extra errors to make sure things add up.
        """
        self._debug = debug
        return self

    def primary_key_config(self, primary_key: Iterable[str], *, natural_primary_key: bool = False):
        """ Customize the primary key settings.

        You will only need this with natural primary keys, or other really special cases.
        """
        self._primary_key = tuple(primary_key)
        self._natural_primary_key = natural_primary_key
        return self

    def create_or_update_config(self, CreateOrUpdateSchema: PydanticModelT):
        """ Enable create_or_update() and provide a schema for it """
        self.CreateOrUpdateInputSchema = CreateOrUpdateSchema
        self.CreateOrUpdateResponseSchema = self.CreateResponseSchema
        return self

    def exclude_config(self, *,
                       create: Iterable[str] = (),
                       update: Iterable[str] = (),
                       create_and_update: Iterable[str] = ()):
        """ Specify additional fields to be excluded when doing create() an update() -- in addition to other exclusions.

        Can be called multiple times.

        Args:
            create: Fields to exclude when doing create()
            update: Fields to exclude when doing update()
            create_and_update: Fields to exclude both when doing create() or update()
        """
        self._extra_exclude_on_create |= frozenset(create) | frozenset(create_and_update)
        self._extra_exclude_on_update |= frozenset(update) | frozenset(create_and_update)
        return self

    def field_names_config(self, *,
                           ro_fields: Iterable[str] = (),
                           ro_relations: Iterable[str] = (),
                           const_fields: Iterable[str] = (),
                           rw_fields: Iterable[str] = (),
                           rw_relations: Iterable[str] = (),
                           ):
        """ Configure the behavior of individual fields

        No error is given if a field is marked as read-only, const, or ignored, and a value is provided by the UI.
        Why? Because the UI would often load values from the server and just post them back.
        It would have been a pain in the neck to remove those keys from the data before sending.
        Therefore, CrudBase will just ignore them.

        NOTE: if you use this method, you're supposed to mention every field, telling whether it's `ro` or `rw`.
        If a field should be ignored, use exclude_config()

        Args:
            ro_fields: Read-only field names. They can never be set or modified.
                Example: primary key, fields that get automatic values, etc
            ro_relations: Read-only field names. Same thing, but with a visual emphasis on relationships.
            rw_fields: Writable fields. They can always be set or modified.
                Example: all sorts of user input data
            rw_relations: Writable fields. Same thing, but with a visual emphasis on relationships.
            const_fields: Constant fields. They can only be provided when creating, but not when updating.
                Example: permanent data
        """
        # Make sure that `_exclude_on_update` did not cache anything yet.
        # Caching is done by putting the cached value inside the object's __dict__
        assert type(self)._exclude_on_update.attrname not in self.__dict__

        # remember
        self._field_names_configured = True
        self._ro_fields = frozenset(ro_fields) | frozenset(ro_relations)
        self._rw_fields = frozenset(rw_fields) | frozenset(rw_relations)
        self._const_fields = frozenset(const_fields)
        return self

    def customize(self, **overrides):
        """ Override arbitrary values within this class

        This is low-level customization.
        Use it in case you want, for instance, to have different response schemas for different methods.
        """
        self.__dict__.update(overrides)
        return self

    def test_crud_configuration(self, CrudHandler: Type['crudbase.SimpleCrudBase']):
        """ Test whether everything we've done so far makes sense. Do so in tests.

        Args:
            CrudHandler: Crud handler class
        """
        validate_multiple(self, CrudHandler, [
            validate_all_field_names_are_known,
            validate_every_copied_field_is_known,
            validate_saves_custom_fields,
            validate_saves_custom_fields_handles_every_relationship,
            validate_ro_rw_field_names_cover_whole_model,
            validate_primary_key_is_not_writable,
        ])
        return self

    # endregion

    # region Usage

    # Tell which options are configured

    @property
    def _create_or_update_enabled(self):
        """ Check: create_or_update() schemas configured and enabled? """
        return self.CreateOrUpdateInputSchema and self.CreateOrUpdateResponseSchema

    # Methods that help use these settings

    def _primary_key_provided(self, provided_names: Iterable[str]):
        """ Check whether the input contains a primary key

        This is necessary to determine whether we will create() or update()
        """
        return set(provided_names) >= set(self._primary_key)

    @cached_property
    def _exclude_on_create(self) -> FrozenSet[str]:
        """ Get the list of fields to exclude when creating an instance """
        exclude = set()

        # If ro/rw fields were configured
        if self._field_names_configured:
            exclude |= self._ro_fields
        # If not, only exclude the PKf
        else:
            exclude |= set() if self._natural_primary_key else set(self._primary_key)

        # Done
        return frozenset(exclude) | self._extra_exclude_on_create

    @cached_property
    def _exclude_on_update(self) -> FrozenSet[str]:
        """ Get the list of fields to exclude when updating an instance """
        exclude = set()
        exclude |= self._exclude_on_create
        exclude |= self._extra_exclude_on_update

        # If ro/rw fields were configured
        if self._field_names_configured:
            exclude |= self._const_fields
        # if not, there's nothing to add: _exclude_on_create() has already done everything
        else:
            pass

        return frozenset(exclude)

    # endregion


# region Validate


def validate_multiple(settings: CrudSettings, crud: type, validators: Iterable[callable]):
    """ Run multiple validators and raise all results together

    This is convenient when multiple validation errors are present. If they were raised immediately,
    you would only get the first one, which is annoying.

    Raises:
        AssertionError, MultipleAssertionErrors
    """
    # Collect errors
    errors = []
    for validator in validators:
        try:
            validator(settings, crud)
        except AssertionError as e:
            errors.append(e)

    # Report errors
    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise MultipleAssertionErrors(*errors)


def validate_all_field_names_are_known(settings: CrudSettings, CrudHandler: type):
    """ Validate: all fields mentioned anywhere are actually known

    This is to prevent typos: you may think you have excluded a field properly, but made a typo.
    This test makes sure that every field you exclude is actually known to a model.
    """
    mentioned_fields = set()
    mentioned_fields |= set(settings._primary_key)
    mentioned_fields |= set(settings._extra_exclude_on_create)
    mentioned_fields |= set(settings._extra_exclude_on_update)
    if settings._field_names_configured:
        mentioned_fields |= set(settings._ro_fields)
        mentioned_fields |= set(settings._rw_fields)
        mentioned_fields |= set(settings._const_fields)

    # known fields
    known_fields = set(dir(settings.Model))

    # Unknown fields
    unknown_fields = mentioned_fields - known_fields
    assert not unknown_fields, (
        f'Unknown fields provided to crudsettings: {unknown_fields!r}. '
        f'Probably, there has been a typo, or changes were made to the model.'
    )


def validate_ro_rw_field_names_cover_whole_model(settings: CrudSettings, CrudHandler: type):
    """ Validate: field_names_config() has covered the whole input model """
    if settings._field_names_configured:
        # Field names
        all_input_fields = _all_input_field_names(settings)
        all_ro_rw_fields = settings._ro_fields | settings._rw_fields | settings._const_fields | settings._extra_exclude_on_update | settings._extra_exclude_on_create

        # Compare
        uncovered_field_names = all_input_fields - all_ro_rw_fields

        # Check
        assert not uncovered_field_names, (
            f'field_names_config() was used, but some fields have not been covered: {uncovered_field_names!r}. '
            f'The main idea of field_names_config() is that you have to specify for every field whether it is writable or not.'
        )


def validate_saves_custom_fields(settings: CrudSettings, CrudHandler: type):
    """ Validate: @saves_custom_fields() is properly configured """
    known_input_fields = _all_input_field_names(settings)
    custom_field_names = crudbase.saves_custom_fields.all_field_names_from(CrudHandler)
    unknown_fields = custom_field_names - known_input_fields
    assert not unknown_fields, (
        f'Unknown fields in @saves_custom_fields(): {unknown_fields}. '
        f'This is likely because you have a typo in the arguments provided to the decorator, or had a change in your models.'
    )


def validate_saves_custom_fields_handles_every_relationship(settings: CrudSettings, CrudHandler: type):
    """ Validate: every writable relationship is covered by @saves_custom_fields()  """
    # Only validate if going to do saving
    if any([settings.CreateInputSchema, settings.UpdateInputSchema, settings.CreateOrUpdateInputSchema]):
        # Get all relationships
        relationship_names = set(sa2schema.sa_model_info(settings.Model, types=sa2schema.AttributeType.RELATIONSHIP))

        # Get relationship names covered by somethind
        handled_relationships = set()
        # @saves_custom_fields
        handled_relationships |= crudbase.saves_custom_fields.all_field_names_from(CrudHandler)
        # exclude
        handled_relationships |= settings._exclude_on_create | settings._exclude_on_update

        # Compare
        unhandled_relationships = relationship_names - handled_relationships
        assert not unhandled_relationships, (
            f'The following relationships have not been handled by @saves_custom_fields(): '
            f'{unhandled_relationships!r}. '
            f'Either implement save handlers with @saves_custom_fields(), '
            f'or exclude them explicitly (`ro_relations`).'
        )


def validate_every_copied_field_is_known(settings: CrudSettings, CrudHandler: type):
    """ Validate: every un-excluded field is known to a model

    This is to make sure that there are no custom fields in Pydantic models that will be copied over to Sqlalchemy models.
    """
    # Collect fields that are ok to be present in the input
    known_fields = set()
    # Model fields
    known_fields.update(dir(settings.Model))
    # excluded fields: are okay because they are excluded for a reason
    known_fields.update(settings._exclude_on_update)
    known_fields.update(settings._exclude_on_create)
    # @saves_custom_fields fields
    known_fields.update(crudbase.saves_custom_fields.all_field_names_from(CrudHandler))

    # Collect fields mentioned in schemas
    pydantic_schemas = _all_input_pydantic_schemas_from_crudsettings(settings)

    # Collect unknown fields
    unknown_fields = []
    for schema_name, schema in pydantic_schemas.items():
        unknown_fields.extend((
            f'{schema_name}.{field_name}'
            for field_name in set(schema.__fields__) - known_fields
        ))

    # Report
    assert not unknown_fields, (
        f"Unknown fields used in schemas: {', '.join(unknown_fields)}. "
        f"This may mean that your Pydantic schemas contain custom fields that you have forgotten to exclude or process manually with @saves_custom_fields"
    )


def validate_primary_key_is_not_writable(settings: CrudSettings, CrudHandler: type):
    """ Validate: primary key should not be writable, unless `natural_primary_key` is selected """
    pk_fields = set(settings._primary_key)
    # Two modes
    # If `natural_primary_key` is not set, create() and update() cannot contain any PK
    if not settings._natural_primary_key:
        # create() cannot contain the primary key. Otherwise, users would be able to book custom keys, which is not right
        # update() can contain the primary key, but only for searching purposes. It must be excluded anyway, because in this function,
        # we're only interested in the fields that get saved to the DB
        fields_create = pk_fields - settings._exclude_on_create
        fields_update = pk_fields - settings._exclude_on_update
        assert not fields_create and not fields_update, (
            f"Primary key fields not excluded for create(): {fields_create!r}, for update(): {fields_update}. "
            f"A user can choose any primary key they want. "
            f"If this is the desired behavior, set `crudsettings.primary_key_config(..., natural_primary_key=True)"
        )
    # If `natural_primary_key` is set, the requirement is the opposite: primary key must be fully present
    else:
        fields_create = pk_fields & settings._exclude_on_create
        fields_update = pk_fields & settings._exclude_on_update
        assert not fields_create and not fields_update, (
            f"Natural primary key fields are excluded for create(): {fields_create!r}, for update(): {fields_update}. "
            f"You have set `natural_primary_key=True`, but it does not make sense if primary key fields are being excluded."
        )


def _all_input_pydantic_schemas_from_crudsettings(settings: CrudSettings) -> Mapping[str, PydanticModelT]:
    """ All input schemas """
    schemas = {
        'CreateInputSchema': settings.CreateInputSchema,
        'UpdateInputSchema': settings.UpdateInputSchema,
        'CreateOrUpdateInputSchema': settings.CreateOrUpdateInputSchema,
    }
    return {
        name: schema
        for name, schema in schemas.items()
        if schema is not None
    }


def _all_input_field_names(settings: CrudSettings) -> Set[str]:
    return set().union(*[
        set(schema.__fields__)
        for schema in _all_input_pydantic_schemas_from_crudsettings(settings).values()
    ])


class MultipleAssertionErrors(AssertionError):
    errors: Tuple[AssertionError]

    def __init__(self, *errors: AssertionError):
        self.errors = errors

        # Prepare the message
        bullet = '* ' if len(errors) > 1 else ''
        msg = '\n'.join(f'{bullet}{e}' for e in errors)

        # Done
        super().__init__(msg)


# endregion



from __future__ import annotations

from dataclasses import dataclass

from typing import Optional, Callable, Dict, Type


@dataclass
class DocumentedFunction:
    """ All documentation for one particular function """
    # Doc: the function itself
    function: Optional[FunctionDoc]

    # Doc: every parameter
    parameters: Dict[str, ParameterDoc]

    # Doc: its result
    result: Optional[ResultDoc]

    # Doc: every error
    errors: Dict[Type[BaseException], ErrorDoc]

    # Doc: deprecation note
    deprecated: Optional[DeprecatedDoc]

    def merge(self, another: DocumentedFunction):
        """ Merge the documentation from another marker into this one """
        # Dicts: update()
        self.parameters.update(another.parameters)
        self.errors.update(another.errors)

        # Dataclasses: merge()
        self.function = self.function.merge(another.function) if self.function else another.function
        self.result = self.result.merge(another.result) if self.result else another.result
        self.deprecated = self.deprecated.merge(another.deprecated) if self.deprecated else another.deprecated

    def parse_from_function_docstring(self, func: Callable):
        """ Parse the function's documentation from its docstring """
        # Optional import of the `docstring_parser` library
        from docstring_parser import (
            DocstringParam,
            DocstringReturns,
            DocstringRaises,
            DocstringDeprecated,
        )
        from docstring_parser.google import parse

        # Parse the docstring
        parsed_docstring = parse(func.__doc__)

        # Parse: title & description
        if parsed_docstring.short_description or parsed_docstring.long_description:
            self.add_doc_for_function(
                summary=parsed_docstring.short_description,
                description=parsed_docstring.long_description or None,
            )

        # Parse: docstring sections
        for section in parsed_docstring.meta:
            # Parse: parameters
            if isinstance(section, DocstringParam):
                summary, _, description = section.description.partition('\n')
                self.add_doc_for_parameter(section.arg_name, summary, description or None)
            # Parse: return value
            elif isinstance(section, DocstringReturns):
                self.add_doc_for_result(section.description, None)
            # Parse: deprecation marker
            elif isinstance(section, DocstringDeprecated):
                self.add_doc_deprecated(section.version, section.description, None)
            # Parse: errors
            elif isinstance(section, DocstringRaises):
                self.add_doc_exception_name(func, section.type_name.strip(), section.description, None)
            # Parse: unknown sections
            else:
                self.add_doc_for_function(summary='', description=''.join(section.args) + ':\n' + section.description)

    def add_doc_for_function(self, summary: str, description: Optional[str]):
        """ Add function documentation """
        if not self.function:
            self.function = FunctionDoc('', None)
        self.function.merge(FunctionDoc(
            summary,
            description or None
        ))

    def add_doc_for_parameter(self, name: str, summary: str, description: Optional[str]):
        """ Add documentation for a parameter """
        self.parameters[name] = ParameterDoc(
            name=name,
            summary=summary,
            description=description or None,
        )

    def add_doc_for_result(self, summary: str, description: Optional[str]):
        """ Add documentation for the return value """
        if not self.result:
            self.result = ResultDoc('', None)
        self.result.merge(ResultDoc(
            summary=summary,
            description=description or None,
        ))

    def add_doc_deprecated(self, version: str, summary: str, description: Optional[str]):
        """ Add documentation for a deprecation """
        if not self.deprecated:
            self.deprecated = DeprecatedDoc('', None, '')
        self.deprecated.merge(DeprecatedDoc(
            version=version,
            summary=summary,
            description=description or None,
        ))

    def add_doc_exception(self, error_cls: Type[BaseException], summary: str, description: Optional[str]):
        """ Add documentation for an exception type """
        self.errors[error_cls] = ErrorDoc(
            error=error_cls,
            summary=summary,
            description=description,
        )

    def add_doc_exception_name(self, func: Callable, error_name: str, summary: str, description: Optional[str]):
        """ Add documentation for an exception by name """
        # A globally available exception type?
        if error_name in __builtins__:
            return __builtins__[error_name]

        # Resolve the error type using function's global namespace
        try:
            error_cls = find_object_in_namespace(func.__globals__, error_name)
        except (KeyError, AttributeError) as e:
            raise ValueError(
                f"The docstring for function {func} references an error named {error_name!r}, "
                f"but it cannot be found in the function's globals. Please make sure its available "
                f"either by its name (e.g. `E_FAIL`), or by reference (`exc.E_FAIL`). "
            ) from e

        # Document it
        self.add_doc_exception(error_cls, summary, description)


@dataclass
class DocBase:
    # A short, one-line summary
    summary: str

    # A longer description
    description: Optional[str]

    def merge(self, another: Optional[DocBase]) -> DocBase:
        if another:
            self.summary = (self.summary or '') + (another.summary or '')
            self.description = ((self.description or '') + (another.description or '')) or None
        return self


@dataclass
class FunctionDoc(DocBase):
    """ Documentation for a function """


@dataclass
class ParameterDoc(DocBase):
    """ Documentation for a parameter """
    # Parameter name
    name: str


@dataclass
class ResultDoc(DocBase):
    """ Documentation for the return value """


@dataclass
class ErrorDoc(DocBase):
    """ Documentation for a known exception """
    error: Type[BaseException]


@dataclass
class DeprecatedDoc(DocBase):
    """ Documentation for a function deprecation """
    # When was it deprecated
    version: str


def find_object_in_namespace(namespace: dict, reference: str):
    """ Given a namespace (like globals()), get the object referenced by `reference`

    Example:
        find_object_in_namespace( func.__globals__, 'errors.E_UNEXPECTED_ERROR' )
    """
    # Support "object.name" attribute access
    name, _, attribute = reference.partition('.')

    # Get the object itself
    object = namespace[name]

    # Get the attribute, recursively
    while True:
        attribute, _, next_attribute = attribute.partition('.')
        object = getattr(object, attribute)
        if not next_attribute:
            break

    # Done
    return object



import mongosql
import pydantic as pd
import sa2schema
import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.orm.attributes
import sqlalchemy.orm.base
from nplus1loader import raiseload_all
from sqlalchemy.orm import load_only, defaultload
from typing import Union, Iterable, ClassVar, Generic, Type

from apiens.views.crud import CrudBase, UserFilterValue, InstanceDict
from apiens.views.crud.crudbase import SAInstanceT
from apiens.views.mongoquery_crud.defs import QueryObject
from apiens.views.mongoquery_crud.mongocrud_settings import MongoCrudSettings


# Generic[SAInstanceT]: your SqlAlchemy model
class MongoCrudBase(CrudBase[SAInstanceT, dict], Generic[SAInstanceT]):
    """ MongoSQL-enabled CRUD

    Additional features:

    * Crud operations load that absolute minimum number of fields
    * MongoSQL performs an additional SELECT query with user-defined customizations
    * MongoSQL tries to make this query as fast as possible
    * Returns plain dicts, not Pydantic models
    """

    crudsettings: ClassVar[MongoCrudSettings]

    # A QueryObject from the request
    query_object: QueryObject

    def __init__(self, ssn: sa.orm.Session, *, query_object: QueryObject = None, **kwargs):
        super().__init__(ssn, **kwargs)
        self.query_object = query_object

    __slots__ = 'query_object',

    # region CRUD methods

    def get(self, **kwargs: UserFilterValue) -> dict:
        # Don't use CrudBase.get() because it will load ORM objects, which is suboptimal.
        # Use MongoQuery instead
        return self._result_fetch_one(self.crudsettings.GetResponseSchema, **kwargs)

    def list(self, **kwargs: UserFilterValue) -> Iterable[dict]:
        return self._result_fetch_many(self.crudsettings.ListResponseSchema, **kwargs)

    def count(self, **kwargs: UserFilterValue) -> int:
        return self._result_fetch_count(**kwargs)

    def create(self, input: Union[InstanceDict, pd.BaseModel]) -> dict:
        # create()
        instance: SAInstanceT = super().create(input)

        # re-load the instance with the user's projection
        result = self._result_refetch_instance(self.crudsettings.CreateResponseSchema, instance)

        # Done
        return result

    def update(self, input: Union[InstanceDict, pd.BaseModel], **kwargs: UserFilterValue) -> dict:
        # update() the instance
        instance: SAInstanceT = super().update(input, **kwargs)

        # re-load the instance with the user's projection
        result = self._result_refetch_instance(self.crudsettings.UpdateResponseSchema, instance)

        # Done
        return result

    def delete(self, **kwargs: UserFilterValue) -> dict:
        # load the instance with a custom projection before it is removed!
        # This is the only way we can fetch all those relationships and everything
        result = self._result_fetch_one(self.crudsettings.DeleteResponseSchema, **kwargs)

        # delete()
        super().delete(**kwargs)

        # Done
        return result

    # endregion

    # region MongoSql Query

    def _result_refetch_instance(self, response_schema: Type[pd.BaseModel], instance: SAInstanceT) -> dict:
        """ Fetch the instance again through a MongoQuery and return an object """
        mq = self._mongoquery_instance_by_primary_key(instance)
        result = self._one_result_from_mongoquery(mq, response_schema)
        return result

    def _result_fetch_one(self, /, response_schema: Type[pd.BaseModel], **kwargs) -> dict:
        """ Fetch one object through a MongoQuery """
        mq = self._mongoquery(*self._filter1(**{**self.kwargs, **kwargs}))
        result = self._one_result_from_mongoquery(mq, response_schema)
        return result

    def _result_fetch_many(self, /, response_schema: Type[pd.BaseModel], **kwargs) -> Iterable[dict]:
        """ Fetch many objects through a MongoQuery """
        mq = self._mongoquery(*self._filter(**{**self.kwargs, **kwargs}))
        results = self._many_results_from_mongoquery(mq, response_schema)
        return results

    def _result_fetch_count(self, **kwargs) -> int:
        """ Fetch the number of matching objects """
        q: sa.orm.Query = self._query(*self._filter(**{**self.kwargs, **kwargs}))
        mq = self.crudsettings._mongoquery(q, {
            **self.query_object,
            'count': 1,  # override it with a count
        })
        q = mq.end()
        return q.scalar()

    def _mongoquery(self, *filter, **filter_by) -> mongosql.MongoQuery:
        """ Start a MongoQuery to load instance[s]

        This method is used in get() an list(): using the filter from the user-supplied query
        """
        query = super()._query(*filter, **filter_by)
        return self.crudsettings._mongoquery(query, self.query_object)

    # Override the CrudBase output and let the instance pass through
    # We do it like this because we don't need that instance at all.
    # Instead, we'll make another query and fetch it with MongoSQL using the custom projection from the QueryObject

    def _instance_output(self, instance: SAInstanceT, schema: Type[pd.BaseModel]) -> SAInstanceT:
        return instance

    # endregion

    # region Query

    def _query(self, *filter, **filter_by) -> Union[sa.orm.Query, Iterable[SAInstanceT]]:
        # Prepare the query
        query = super()._query(*filter, **filter_by)

        # In MongoCrudBase, self._query() only serves the update() and delete() operations.
        # The update() and delete() don't usually need many fields.
        # Put a load_only() on it, so that it does not load more that necessary.
        query = query.options(
            load_only(
                # undefer the primary key and all the foreign keys
                *primary_and_foreign_keys_from(self.crudsettings.Model),
                # custom fields
                *self.crudsettings._load_only,
            ),
            # Be strict
            raiseload_all('*') if self.crudsettings._debug else defaultload([]),
        )

        # Done
        return query

    # endregion

    # region Helpers

    def _mongoquery_instance_by_primary_key(self, instance: SAInstanceT):
        """ Start a MongoQuery to load a known instance by primary key """
        # Extract the primary key and build the condition
        filter = self._filter_primary_key(
            # Use the instance dict as the `kwargs`.
            # This will work because `kwargs` is expected to contain attribute name-value pairs anyway
            kwargs=sa.orm.base.instance_dict(instance)
        )

        # mongoquery
        return self._mongoquery(*filter)

    def _one_result_from_mongoquery(self, mq: mongosql.MongoQuery, response_schema: Type[pd.BaseModel]) -> dict:
        """ Prepare a singular result from a MongoQuery """
        # Query
        query = mq.end()
        mq_instance = query.one()

        # Pluck
        result = mq.pluck_instance(mq_instance)

        # Optionally, in debug mode, validate the response model
        if self.crudsettings._debug:
            response_schema.parse_obj(result)

        # Done
        return result

    def _many_results_from_mongoquery(self, mq: mongosql.MongoQuery, response_schema: Type[pd.BaseModel]) -> Iterable[dict]:
        """ Prepare a plural result from a MongoQuery """
        mq_instances: Union[sa.orm.Query, Iterable[SAInstanceT]] = mq.end()

        # Pluck
        results = (
            mq.pluck_instance(mq_instance)
            for mq_instance in mq_instances
        )

        # Validate
        if self.crudsettings._debug:
            results = list(results)  # otherwise the generator would be exhausted
            for item in results:
                response_schema.parse_obj(item)

        # Done
        return results

    # endregion


def primary_and_foreign_keys_from(model: type) -> Iterable[sa.orm.attributes.InstrumentedAttribute]:
    """ Get model attributes: primary and foreign keys """
    columns_info = sa2schema.sa_model_info(model, types=sa2schema.AttributeType.COLUMN)
    attr_info: sa2schema.info.ColumnInfo
    return [
        attr_info.attribute
        for attr_name, attr_info in columns_info.items()
        if attr_info.primary_key or attr_info.foreign_key
    ]



from typing import Mapping, Any


# Values for an SqlAlchemy instance, in the form of a dictionary. Possibly, partial.
InstanceDict = Mapping[str, Any]

# User-supplied values for filtering objects
UserFilterValue = Any


# Sugar. Marks a field that will be automatically initialized.
AUTOMATIC = None



""" Class-based views for Create, Read, Update, Delete operations """

from . import crud_signals
from .defs import AUTOMATIC, InstanceDict, UserFilterValue
from .crudbase import SimpleCrudBase, CrudBase
from .crudbase import saves_custom_fields
from .crud_settings import CrudSettings
from .instance_history_proxy import InstanceHistoryProxy, get_history_proxy_for_instance

try:
    from .pg_integrity_error import extract_postgres_unique_violation_columns
except ImportError as e:
    if e.name != 'psycopg2':
        raise



""" Lighweight dependency injection """

from .injector import Injector
from .markers import kwargs, depends, signature

from .const import InjectFlags
from .defs import InjectionToken, Resolvable, Provider, Dependency, Dependency
from .exc import BaseInjectorError, NoProviderError, ClosedInjectorError

from .markers import resolvable_marker

# Shortcuts
DEFAULT = InjectFlags.DEFAULT
SELF = InjectFlags.SELF
SKIP_SELF = InjectFlags.SKIP_SELF
OPTIONAL = InjectFlags.OPTIONAL



from contextlib import contextmanager

import pydantic as pd
import sqlalchemy as sa
import sqlalchemy.orm
from typing import TypeVar, Generic, Iterable, Union, Mapping, Any, ClassVar, Set, Sequence, Type, ContextManager, Optional

from apiens.util import decomarker
from . import crud_signals
from .crud_settings import CrudSettings
from .defs import UserFilterValue, InstanceDict
from .instance_history_proxy import get_history_proxy_for_instance


# SqlAlchemy instance
SAInstanceT = TypeVar("SAInstanceT")

# Response object
ResponseValueT = TypeVar("ResponseValueT")

# Crud Handler
CrudHandlerT = TypeVar('CrudHandlerT')


# Generic[SAInstanceT]: your SqlAlchemy model
class SimpleCrudBase(Generic[SAInstanceT]):
    """ CRUD Handler: basic business-logic layer for your SqlAlchemy instances.

    Main features:

    * Independent from any web framework: works purely with data
    * Only implements operations on instances themselves. Does not save anything to a Session.
    * Uses Pydantic models for input data validation
    * Uses Pydantic models for output data modification
    * Supports any means of filtering/selection using *filter and **filter_by arguments
    * Sends CRUD signals. See `crud_signals.py`
    * Can save relationships (by using @saves_custom_fields)
    * Works with SqlAlchemy Session and actually stores objects into the database
    * Does not commit(). You have to commit() manually

    This object is supposed to be re-initialized for every request.

    For a more thorough implementation that is able to actually save objects, see `CrudBase` class.
    """

    # CRUD settings: the static container for all the configuration of this CRUD handler.
    crudsettings: ClassVar[CrudSettings]

    # The database session to use
    ssn: sqlalchemy.orm.Session

    def __init__(self, ssn: sa.orm.Session):
        """ Initialize the Crud handler for one request

        Args:
            ssn: The Session to work with. This implementation only uses it for loading.
        """
        self.ssn = ssn

    __slots__ = 'ssn'

    # Use deep copying for historical `prev` objects?
    # Set to `True` if your Crud handler requires accessing the historical value of a mutable field (dict, etc)
    # Technically, it will use InstanceHistoryProxy(copy=True)
    COPY_INSTANCE_HISTORY: bool = False


    # region CRUD Operation implementations

    # Genus: _*_instance()
    # input: filter, Pydantic model
    # output: SqlAlchemy instance
    # Purpose: implement CRUD on instance level

    # These methods implement the basic features of CRUD operations: get the input, apply it, yield results.
    # Every method receives *filter and **filter_by arguments that let you apply custom filtering.
    # They all return SqlAlchemy models, or Query[SAInstanceT] iterables
    # They do not flush() nor commit()
    # These are the basic building blocks for your application.

    def _get_instance(self, *filter: sa.sql.expression.BinaryExpression, **filter_by: Mapping[str, Any]) -> SAInstanceT:
        """ get() method: load one instance

        Args:
            *filter: SqlAlchemy filter() expressions (Model.field == value)
            **filter_by: SqlAlchemy filter_by() expressions ( { 'field_name': value } )

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
        """
        # Raises: sqlalchemy.orm.exc.NoResultFound
        # Raises: sqlalchemy.orm.exc.MultipleResultsFound
        return self._query(*filter, **filter_by).one()

    def _list_instances(self, *filter: sa.sql.expression.BinaryExpression, **filter_by: Mapping[str, Any]) -> Union[sa.orm.Query, Iterable[SAInstanceT]]:
        """ list() method: load multiple instances

        This method will ssn.query() them and return

        Returns:
            Query[SAInstanceT] that can be iterated for models
        """
        return self._query(*filter, **filter_by)

    def _count_instances(self, *filter: sa.sql.expression.BinaryExpression, **filter_by: Mapping[str, Any]) -> int:
        """ count() method: filter instances and give their count """
        return self._query(*filter, **filter_by).count()

    def _create_instance(self, input: pd.BaseModel) -> SAInstanceT:
        """ create() method: create an instance

        This method will Model(**input.dict()) , do @saves_custom_fields(), but won't ssn.add() it
        Session operations are out of scope.

        Args:
            input: Input from the user.
                It is a validated Pydantic model and it will be used to create an SqlAlchemy model.
                NOTE: it has to be a Pydantic model.
                CrudBase.create() can also accept a dictionary, but it's a Pydantic model all the way down.

        Raises:
            sa.exc.IntegrityError
        """
        # Create
        instance = self._create_instance_from_input(input, exclude=self.crudsettings._exclude_on_create)

        # Hook
        self._instance_hook_presave(new=instance, prev=None)

        # Signals
        crud_signals.on_create_prepared.send(type(self), crud=self, new=instance)
        crud_signals.on_save_prepared.send(type(self), crud=self, new=instance, prev=None, action='create')

        # Done
        return instance

    def _update_instance(self, input: pd.BaseModel, *filter, **filter_by) -> SAInstanceT:
        """ update() method: load an existing instance and modify it

        This method will ssn.query() it, setattr() changes, do @saves_custom_fields(), but won't ssn.flush() it.
        Session operations are out of scope.

        Args:
            input: Input from the user: a validated Pydantic model.

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
            sa.exc.IntegrityError
        """
        # Raises: sqlalchemy.orm.exc.NoResultFound
        # Raises: sqlalchemy.orm.exc.MultipleResultsFound
        instance = self._query(*filter, **filter_by).one()
        prev = get_history_proxy_for_instance(instance, copy=self.COPY_INSTANCE_HISTORY)

        # Update it
        instance = self._update_instance_from_input(instance, input, exclude=self.crudsettings._exclude_on_update)

        # Hook
        self._instance_hook_presave(new=instance, prev=prev)

        # Signals
        crud_signals.on_update_prepared.send(type(self), crud=self, new=instance, prev=prev)
        crud_signals.on_save_prepared.send(type(self), crud=self, new=instance, prev=prev, action='update')

        # Done
        return instance

    def _delete_instance(self, *filter, **filter_by) -> SAInstanceT:
        """ delete() method: delete an existing instance and return it

        This method will ssn.query() it ... and nothing else.
        Session operations are out of scope.

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
        """
        # Raises: sqlalchemy.orm.exc.NoResultFound
        # Raises: sqlalchemy.orm.exc.MultipleResultsFound
        instance = self._query(*filter, **filter_by).one()

        # Hook
        self._instance_hook_presave(new=None, prev=instance)

        # Signals
        crud_signals.on_delete_prepared.send(type(self), crud=self, prev=instance)
        crud_signals.on_save_prepared.send(type(self), crud=self, new=None, prev=instance, action='delete')

        # Done
        return instance

    def _instance_hook_presave(self, new: Optional[SAInstanceT] = None, prev: Optional[SAInstanceT] = None):
        """ A hook called when an instance is going to be saved. Pre-flush, before signals.

        This hook is called before_flush:
        * on create(), after an instance has been constructed from the input
        * on update(), after an instance has been modified from the input
        * on delete(), after it has been loaded, before it's marked for deletion

        In all cases, this hook is called before any other signal is sent.
        Use it for fine-tuning created objects.

        Features:
        * The instance has just been modified
        * No other signal handlers have been fired
        * It is before-flush, so all the changes are still visible

        Here's how you know what's going on:
        * new and not old: create()
        * new and old: update()
        * not new and old: delete()

        Args:
            new: The instance being saved. `None` when delete()
            prev: The unmodified version of the instance. `None` when create()
        """

    # region Low-Level: assign values to attributes

    # Genus: _*_instance_from_input()
    # Input: Pydantic model
    # Output: SqlAlchemy Model, created/modified
    # Purpose: apply input model to actual SqlAlchemy model

    # Methods that simply create/modify an instance from an input
    # Create/Update an instance using the input model

    def _create_instance_from_input(self, input: pd.BaseModel, *, exclude: Set[str]) -> SAInstanceT:
        """ Given an `input` model, create a new SA Model, do @saves_custom_fields(), return """
        input_dict = input.dict(exclude_unset=True, exclude=exclude)

        # Save, wrapped with @saves_custom_fields() handler
        custom_fields = saves_custom_fields.pluck_custom_fields(self, input_dict)
        instance = self._create_instance_from_input_dict(input_dict)
        saves_custom_fields.save(self, custom_fields, instance, None)

        return instance

    def _update_instance_from_input(self, instance: SAInstanceT, input: pd.BaseModel, *, exclude: Set[str]) -> SAInstanceT:
        """ Given an `input` and an `instance`, update it, do @saves_custom_fields(), return """
        input_dict = input.dict(exclude_unset=True, exclude=exclude)

        # Save, wrapped with @saves_custom_fields() handler
        custom_fields = saves_custom_fields.pluck_custom_fields(self, input_dict)
        instance = self._update_instance_from_input_dict(instance, input_dict)
        saves_custom_fields.save(self, custom_fields, instance, None)

        return instance

    # Create/Update an instance using the input dict

    def _create_instance_from_input_dict(self, input_dict: dict) -> SAInstanceT:
        """ Given an `input_dict` dict, create a new Model """
        return self.crudsettings.Model(**input_dict)

    def _update_instance_from_input_dict(self, instance: SAInstanceT, input_dict: dict) -> SAInstanceT:
        """ Given an `input_dict` and an `instance`, update it """
        for name, value in input_dict.items():
            setattr(instance, name, value)  # triggers SqlAlchemy change detection logic
        return instance

    # endregion

    # endregion


    # region Low-level CRUD operations

    # These methods implement CRUD operations on SqlAlchemy instances.
    # This is unlike the get()/list()/create()/update()/delete() operations that work with Pydantic models.



    # region Session Saving Operations

    # Genus: _session_*_instance() methods
    # input: filter, Pydantic model
    # output: SqlAlchemy instance
    # They call: _*_instance()
    # Purpose: use Session to save things

    # These methods implement the SqlAlchemy Session saving logic

    def _session_create_instance(self, input: pd.BaseModel) -> SAInstanceT:
        """ Session support for create(): create an instance and flush() it """
        # Create
        instance = self._create_instance(input)

        # Flush
        self.ssn.add(instance)
        self.ssn.flush()

        # Signals
        crud_signals.on_create.send(type(self), crud=self, new=instance)
        crud_signals.on_save.send(type(self), crud=self, new=instance, prev=None, action='create')

        # Done
        return instance

    def _session_update_instance(self, input: pd.BaseModel, *filter, **filter_by) -> SAInstanceT:
        """ Session support for update(): load an instance, modify it, flush() it """
        # Update
        instance = self._update_instance(input, *filter, **filter_by)
        prev = get_history_proxy_for_instance(instance)  # returns the proxy that has already been initialized by _update_instance()

        # Flush
        self.ssn.add(instance)
        self.ssn.flush()

        # Signals
        crud_signals.on_update.send(type(self), crud=self, new=instance, prev=prev)
        crud_signals.on_save.send(type(self), crud=self, new=instance, prev=prev, action='update')

        # Done
        return instance

    def _session_delete_instance(self, *filter, **filter_by) -> SAInstanceT:
        """ Session support for delete(): load an instance, Session.delete() it, flush it  """
        # Delete
        instance = self._delete_instance(*filter, **filter_by)

        # Workaround: remove all possible lazy loads and raiseloads. raiseload() is especially harmful here:
        # SqlAlchemy will do cascade deletions, and it will need the values of foreign key fields.
        # See: https://github.com/sqlalchemy/sqlalchemy/issues/5398
        self.ssn.refresh(instance)

        # Flush
        self.ssn.delete(instance)
        self.ssn.flush()

        # Signals
        crud_signals.on_delete.send(type(self), crud=self, prev=instance)
        crud_signals.on_save.send(type(self), crud=self, new=None, prev=instance, action='delete')

        # Done
        return instance

    # endregion

    # endregion


    # region Querying

    def _query(self, *filter, **filter_by) -> Union[sa.orm.Query, Iterable[SAInstanceT]]:
        """ Start a Query for loading instances of our Model

        Args:
            *filter: SqlAlchemy filter() expressions (Model.field == value)
            **filter_by: SqlAlchemy filter_by() expressions ( { 'field_name': value } )
        """
        # Query
        q = self.ssn.query(self.crudsettings.Model)

        # We have to apply our filtering in advance, because later on, someone (like MongoSQL) may put on a LIMIT,
        # and no filter() will be possible anymore
        if filter:
            q = q.filter(*filter)
        if filter_by:
            q = q.filter_by(**filter_by)

        # Done
        return q

    @contextmanager
    def transaction(self: Type[CrudHandlerT]) -> ContextManager[CrudHandlerT]:
        """ Wrap a section of code into a Crud transaction. commit() if everything goes fine; rollback() if not

        Example:
            with UserCrud(ssn).transaction() as crud:
                user = crud.update(user, id=id)
                return {'user': user}
        """
        try:
            yield self
        except Exception:
            self.rollback()
            raise
        else:
            self.commit()

    def commit(self) -> sa.orm.Session:
        """ Actually commit the changes.

        Note that no other place in this class performs a commit!
        You have to call it when you're done with this CRUD object.
        """
        # Send signals and commit
        crud_signals.on_commit_before.send(type(self), crud=self)
        self.ssn.commit()
        crud_signals.on_commit_after.send(type(self), crud=self)
        return self.ssn

    def rollback(self) -> sa.orm.Session:
        """ Send signals and Session.rollback() """
        # Send signals and rollback
        crud_signals.on_rollback_before.send(type(self), crud=self)
        self.ssn.rollback()
        crud_signals.on_rollback_after.send(type(self), crud=self)
        return self.ssn

    # endregion


# Generic[SAInstanceT]: your SqlAlchemy model
# ResponseValueT[ResponseValueT]: the type your custom _instance_output() returns.
#   That is, the final output type for your crud methods.
class CrudBase(SimpleCrudBase[SAInstanceT], Generic[SAInstanceT, ResponseValueT]):
    """ Feature-complete Crud Handler for your application's business logic

    In addition to SimpleCrudBase, provides:

    * Public CRUD methods: get(), list(), create(), update(), delete(), etc
    * Receives **kwargs key-value pairs that the CRUD implementation is supposed to interpret
    * Customizable _filter() to specify which objects fall into the scope of this handler
    * Customizable _filter1() to specify how to find one particular entity.
      Initially, geared up to find objects by their primary key

    """

    # Keyword arguments given to the CRUD handler
    kwargs: dict

    def __init__(self, ssn: sa.orm.Session, **kwargs):
        """

        Args:
            **kwargs: Filtering values for filter() for every operation
        """
        super().__init__(ssn)
        self.kwargs = kwargs

    __slots__ = 'kwargs',

    # region Filtering

    # These methods customize how your CRUD operations find objects to work with
    # Implement the _filter() and _filter1() methods

    def _filter(self, **kwargs: UserFilterValue) -> Iterable[sa.sql.expression.BinaryExpression]:
        """ Filter expression for all methods. Override and customize to react to user-supplied `kwargs`

        Args:
            **kwargs: User-supplied input for filtering.
                For instance, these can be the arguments of your view.
        """
        return ()

    def _filter1(self, **kwargs: UserFilterValue) -> Iterable[sa.sql.expression.BinaryExpression]:
        """ Filter expression for get(), update(), delete()

        NOTE: the default implementation already filters by the primary key and users _filter().
        You do not have to worry about those!

        Args:
            **kwargs: User-supplied input that is supposed to identify the object, e.g., by primary key
        """
        # The default implementation: _filter() + primary key filter
        return (
            *self._filter(**kwargs),
            *self._filter_primary_key(kwargs),
        )

    def _filter_primary_key(self, kwargs: Mapping[str, UserFilterValue]) -> Iterable[sa.sql.expression.BinaryExpression]:
        """ Find an instance by its primary key values

        Args:
            kwargs: User-supplied input for filtering. Must contain primary key { name => value } pairs
        """
        return (
            getattr(self.crudsettings.Model, pk_field) == kwargs.get(pk_field, None)
            for pk_field in self.crudsettings._primary_key
        )

    # endregion


    # region Public CRUD operations

    # Genus: Public CRUD: get(), list(), create(), update(), delete(), etc, methods
    # input: **kwargs, Pydantic model or dict
    # output: Pydantic model or dict
    # They call: _session_*_instance()
    # Purpose: top-level CRUD methods that implement the full data cycle

    # The final CRUD operations.
    # They transparently work with Pydantic models both on the input and on the output. SqlAlchemy is not visible.
    # Every method accepts **kwargs: the custom data received from the user for filtering. Typically, view arguments.

    def get(self, **kwargs: UserFilterValue) -> ResponseValueT:
        """ get() method: load a single object

        Args:
            **kwargs: filtering arguments

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
        """
        instance = self._get_instance(*self._filter1(**{**self.kwargs, **kwargs}))
        return self._instance_output(instance, self.crudsettings.GetResponseSchema)

    def list(self, **kwargs: UserFilterValue) -> Iterable[ResponseValueT]:
        """ list() method: load multiple objects using criteria """
        instances = self._list_instances(*self._filter(**{**self.kwargs, **kwargs}))
        return self._instances_output(instances, self.crudsettings.ListResponseSchema)

    def count(self, **kwargs: UserFilterValue) -> int:
        """ count() method: get the number of matching instances """
        n = self._count_instances(*self._filter(**{**self.kwargs, **kwargs}))
        return n

    # NOTE: create() and update() methods receive either a dict or a Pydantic model.
    # In both cases, the data will be validated.
    # Extra fields are ignored.

    def create(self, input: Union[InstanceDict, pd.BaseModel]) -> ResponseValueT:
        """ create() method: create a new object in the database

        Args:
            input: input values for the new object. A Pydantic model or a dict

        Raises:
            sa.exc.IntegrityError
        """
        if not isinstance(input, self.crudsettings.CreateInputSchema):  # CHECKME: isinstance(input, pd.BaseModel) ?
            input = self.crudsettings.CreateInputSchema.parse_obj(input)

        instance = self._session_create_instance(input)
        return self._instance_output(instance, self.crudsettings.CreateResponseSchema)

    def update(self, input: Union[InstanceDict, pd.BaseModel], **kwargs: UserFilterValue) -> ResponseValueT:
        """ update() method: load an existing object from the database and modify it

        Args:
            input: new values for the object. A Pydantic model or a dict.
                If it contains a primary key, it may be used instead of `kwargs`.
            **kwargs: filtering arguments

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
            sa.exc.IntegrityError
        """
        # Input
        if not isinstance(input, self.crudsettings.UpdateInputSchema):  # CHECKME: isinstance(input, pd.BaseModel) ?
            input = self.crudsettings.UpdateInputSchema.parse_obj(input)

        # Pull the primary key from the `input`, if provided
        pk_fields_in_input = set(self.crudsettings._primary_key) & input.__fields_set__
        kwargs.update({pk_field: getattr(input, pk_field)
                       for pk_field in pk_fields_in_input
                       if pk_field not in kwargs})

        # Update
        instance = self._session_update_instance(input, *self._filter1(**{**self.kwargs, **kwargs}))
        return self._instance_output(instance, self.crudsettings.UpdateResponseSchema)

    def delete(self, **kwargs: UserFilterValue) -> ResponseValueT:
        """ delete() method: delete an existing object from the database

        Args:
            **kwargs: filtering arguments

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
        """
        instance = self._session_delete_instance(*self._filter1(**{**self.kwargs, **kwargs}))
        return self._instance_output(instance, self.crudsettings.DeleteResponseSchema)

    def create_or_update(self, input: Union[InstanceDict, pd.BaseModel], **kwargs: UserFilterValue) -> ResponseValueT:
        """ update() an object if a primary key was provided in the `input`, create() it otherwise

        This method may look a little odd, but the default policy of this CrudBase is to disallow customization of primary keys.
        That is, if you try to save User(id=100, ...), it will assume that you are updating an existing user.
        It will not let you choose `id=100` just because you like it: it will try to find such a user, and fail if there is none.

        Args:
            input: an object. If a primary key is set, update() will be performed. If not, create() will be done.

        Raises:
            sa.exc.NoResultFound
            sa.exc.MultipleResultsFound
            sa.exc.IntegrityError
        """
        assert self.crudsettings._create_or_update_enabled  # Must be enabled in the CrudSettings

        # Is there a primary key inside?
        input_field_names = set(input.__fields_set__) if isinstance(input, pd.BaseModel) else set(input)
        pk_provided = self.crudsettings._primary_key_provided(input_field_names)

        # Update
        if not pk_provided:
            return self.create(input)
        else:
            return self.update(input, **kwargs)

    # TODO: _method_create_or_update_many() from MongoSQL 2.x

    # endregion


    # region Output

    # TODO: optionally disable validation for response schemas (only in testing)

    def _instance_output(self, instance: SAInstanceT, schema: Type[pd.BaseModel]) -> ResponseValueT:
        """ Convert an SqlAlchemy instance to the final output value """
        # Use Pydantic to convert it
        return schema.from_orm(instance)

    def _instances_output(self, instances: Iterable[SAInstanceT], schema: Type[pd.BaseModel]) -> Iterable[ResponseValueT]:
        """ Convert an iterable of SqlAlchemy instances into the final output value """
        return (
            self._instance_output(instance, schema)
            for instance in instances
        )

    # endregion


class saves_custom_fields(decomarker):
    """ A decorator that marks a method that customizes the saving of certain fields, e.g. relationships.

    It's not safe to *just* let users save relationships and some other fields.
    With the help of this decorator, they are plucked out of the input and handled manually by your function.

    Example:
        ABSENT = object()

        class UserCrud(CrudBase):
            ...
            @saves_custom_fields('articles')
            def save_articles(self, /, new: User, prev: User = None, *, articles = ABSENT):
                if articles is not ABSENT:
                    ...  # custom article-saving logic

    You can use it to save any attributes that require custom behavior, not just relationships.

    The arguments are:

    * new: The instance that is being created/modified
    * prev: The unmodified versions of this instance (only during update())
    * **fields: the values of the fields that you have asked for

    Note that this method is called always, even if no `**fields` have actually been provided.
    """
    # The list of fields that this method is capable of saving.
    field_names = Sequence[str]

    def __init__(self, *field_names: str):
        """ Decorate a method that can save custom fields

        Args:
            *field_names: List of field names that it handles
        """
        super().__init__()
        self.field_names = field_names

    @classmethod
    def save(cls, crud: SimpleCrudBase, plucked_data: Mapping['saves_custom_fields', dict], new: object, prev: object = None):
        for handler, kwargs in plucked_data.items():
            handler.func(crud, new, prev, **kwargs)

    @classmethod
    def pluck_custom_fields(cls, crud: SimpleCrudBase, input_dict: dict) -> Mapping['saves_custom_fields', dict]:
        """ Given an `input_dict`, pluck all custom fields and stash them

        Args:
            crud: The Crud handler
            input_dict: The input dictionary to pluck the values from

        Returns:
            a mapping that will soon be used to call custom fields' handlers.
            It contains: { saves_custom_fields handelr => method kwargs }
        """
        ret = {}
        remove_fields = set()

        # Go through every handler
        for handler in cls.all_decorated_from(type(crud)):
            # For every handler, collect arguments from the dict
            ret[handler] = {
                # NOTE: if an argument has not been provided, use `ABSENT`
                # We don't use `None` to differentiate from a `None` provided by the user
                field_name: input_dict.get(field_name)
                for field_name in handler.field_names
                if field_name in input_dict
            }

            # Remember the fields to remove
            remove_fields.update(handler.field_names)

        # Remove the fields from the dict
        for remove_field in remove_fields:
            if remove_field in input_dict:
                del input_dict[remove_field]

        # Done
        return ret

    @classmethod
    def all_field_names_from(cls, CrudClass: Type[SimpleCrudBase]) -> Set[str]:
        """ Get the names of all custom fields """
        return set().union(*(
            handler.field_names
            for handler in cls.all_decorated_from(CrudClass)
        ))



from typing import Optional, Any

import yaml  # optional dependency; this module is only usable when it's installed
from fastapi import Query

from apiens.views.mongoquery_crud import QueryObject, ModernQueryObject


def query_object(*,
        select: Optional[str] = Query(
            None,
            title='The list of fields to select.',
            description='Example: `[id, login]` or `{ id: 1, login: 1, tenant: { id: 1 }}`. JSON or YAML.',
        ),
        filter: Optional[str] = Query(
            None,
            title='Filter criteria.',
            description='MongoDB format. Example: `{ age: { $gt: 18 } }`. JSON or YAML.'
        ),
        sort: Optional[str] = Query(
            None,
            title='Sorting order',
            description='List of columns with `+` or `-`. Example: `[ login, ctime- ]`. JSON or YAML.',
        ),
        skip: Optional[int] = Query(
            None,
            title='Pagination. The number of items to skip.'
        ),
        limit: Optional[int] = Query(
            None,
            title='Pagination. The number of items to include.'
        ),
) -> Optional[QueryObject]:
    """ Get the MongoSql Query Object from the request

    Raises:
        ArgumentValueError
    """
    # Empty?
    if not select and not filter and not sort and not skip and not limit:
        return None

    # Parse
    query_object = ModernQueryObject(
        select=_parse_yaml_argument('select', select),
        filter=_parse_yaml_argument('filter', filter),
        sort=_parse_yaml_argument('sort', sort),
        skip=skip,
        limit=limit,
    )

    # Convert
    return convert_query_object_to_old(query_object)


class ArgumentValueError(ValueError):
    """ Query object field parse error """
    def __init__(self, argument_name: str, error: str):
        self.argument_name = argument_name
        super().__init__(error)


def _parse_yaml_argument(name: str, value: Optional[str]) -> Any:
    """ Parse a flattened QueryObject field """
    # None passthrough
    if value is None:
        return None

    # Parse the string
    try:
        return yaml.load(value, Loader=yaml.SafeLoader)
    except yaml.YAMLError as e:
        raise ArgumentValueError(name, str(e))


def convert_query_object_to_old(query_object: ModernQueryObject) -> QueryObject:
    """ Convert a Query Object from the new format into the old format

    Two differences:
    * `project` is now called `select`
    * `join` and `joinf` are deprecated and aren't expected to be used
    """
    query_object = query_object.copy()

    # Rename 'select' to 'project'
    if 'select' in query_object:
        project = query_object['project'] = query_object.pop('select')

        # Convert mixed projection into dict projection
        if isinstance(project, list) and any(isinstance(v, dict) for v in project):
            # List projections support one mixed syntax: [name, name, {name: {...}}]
            query_object['project'] = {name: 1 for name in project if isinstance(name, str)}
            for value in project:
                if isinstance(value, dict):
                    query_object['project'].update(value)

        # Convert nested queries
        project = query_object['project']
        if isinstance(project, dict):
            for name, value in project.items():
                if isinstance(value, dict):
                    project[name] = convert_query_object_to_old(value)

    return query_object


def unconvert_query_object_from_old(query_object: QueryObject) -> ModernQueryObject:
    query_object = query_object.copy()
    if 'project' in query_object:
        query_object['select'] = query_object.pop('project')
    return query_object



""" CRUD signals """

import blinker


# An instance has been prepared for saving
# Sent: before flush, so the instance doesn't yet have its database ids. For instance, the primary key is not known yet
# Sender: CRUD Class
on_create_prepared = blinker.signal('on_create_prepared')  # (sender: Type[CrudBase], *, crud: CrudBase, new: object)
on_update_prepared = blinker.signal('on_update_prepared')  # (sender: Type[CrudBase], *, new: CrudBase, prev: object)
on_delete_prepared = blinker.signal('on_delete_prepared')  # (sender: Type[CrudBase], *, prev: object)
on_save_prepared = blinker.signal('on_save_prepared')  # (sender: Type[CrudBase], *, new: Optional[object], prev: Optional[object], action='create')

# An instance is being saved.
# Sent: after flush, so the instance has already gotten its database ids
# Sender: CRUD Class
on_create = blinker.signal('on_create')  # (sender: Type[CrudBase], *, new: object)
on_update = blinker.signal('on_update')  # (sender: Type[CrudBase], *, new: object, prev: object)
on_delete = blinker.signal('on_delete')  # (sender: Type[CrudBase], *, prev: object) NOTE: it's already deleted; be careful, you can't load anything
on_save = blinker.signal('on_save')  # (sender: Type[CrudBase], *, new: Optional[object], prev: Optional[object], action='create')


# Commit
# Sender: CRUD Class
on_commit_before = blinker.signal('on_commit_before')  # (sender: Type[CrudBase], *, crud: CrudBase)
on_commit_after = blinker.signal('on_commit_after')  # (sender: Type[CrudBase], *, crud: CrudBase)


# Rollback
# Sender: CRUD Class
on_rollback_before = blinker.signal('on_rollback_before')  # (sender: Type[CrudBase], *, crud: CrudBase)
on_rollback_after = blinker.signal('on_rollback_after')  # (sender: Type[CrudBase], *, crud: CrudBase)



from typing import ClassVar, Optional


class BaseApplicationError(Exception):
    """ Base Exception class for your application's exceptions

    It provides rich context both to the user and to the developer.
    Such errors are generally identified by name, which is their class name.

    It has the following special features:

    * Every error has a unique name. The UI will be able to tell one error condition from another and handle it.
      For instance, during signing in, a "E_AUTH_CREDENTIALS" may be handled differently from "E_AUTH_USER_DEACTIVATED".
    * Every error has two text messages:
      The "error" field tells what have gone wrong. It's a negative message.
      The "fixit" field tells what to do to fix it. It's a positive message.
    * Errors contain the "info" field with raw, unformatted, data, associated with the error.
      It is very convenient for the UI to have structured information about what went wrong.
    * Use the `.format()` method to insert the `info` into both the `error` and the `fixit`

    Example:

        class E_NOT_FOUND(BaseApplicationError):
            \""" Object not found \"""
            httpcode = 404
            title = _('Not found')

        raise E_NOT_FOUND.format(
            _('Could not find the {object} by {field}'),
            _('Please make sure you have entered a valid email and try again'),
            object=_('User'),
            field=_('email'),
            email='kolypto@gmail.com',
        )

    Such an error should be reported to the API user as a JSON object:

        {
            "name": 'E_NOT_FOUND',
            "title": 'Not found',
            "error": 'Could not find the User by email',
            "fixit": 'Please make sure you have entered a valid email and try again',
            "info": {
                "object": 'User',
                "field": 'email',
                "email": 'kolypto@gmail.com',
            },
            "debug": {},
        }

    So this is what the UI has:

    * "name" lets them know what went wrong and use a specific handler
    * "error" tells what went wrong
    * "fixit" tells what to do
    * "info" contains structured error information
    * "debug" may contain technical error information
    """

    # Every exception is mapped to an HTTP code
    # https://httpstatuses.com/
    httpcode: ClassVar[int]

    # Generic title of the error class in general
    title: ClassVar[str]

    # Error message: what has gone wrong
    # (message to the user)
    error: str

    # Suggestion: what to do in order to fix it
    # (message to the user)
    fixit: Optional[str]

    # Additional information about the error
    info: dict

    # Debug information about the error.
    # This information is considered to be sensitive and is not reported to the user.
    # It's only available in the server logs.
    # Usage: provide `**info` fields named `debug_*` and they will go here
    debug: dict

    def __init__(self, /, error: str, fixit: str = None, **info):
        """
        Args:
            error: What has gone wrong (negative)
            fixit: What to do in order to fix it (positive)
            info: Additional information.
                Prefix fields with `debug_*` to provide sensitive debugging information
        """
        super().__init__(error)
        self.error = error
        self.fixit = fixit or getattr(self, 'fixit', None)  # get the default from a class-level value, if any
        self.info = info

        # Tell `info` and `debug` apart; remove 'debug_' prefixes
        self.info = {k: v for k, v in info.items() if not k.startswith('debug_')}
        self.debug = {k[6:]: v for k, v in info.items() if k.startswith('debug_')}

        # Special features
        self._response_headers = None

    @classmethod
    def format(cls, /, error: str, fixit: str = None, **info):
        """ Make an exception, format() data into `error` and `fixit` messages

        Example:
            E_NOTFOUND.format(
                _('Cannot find {object}'),
                object=object
            )
        """
        # apply format() to `error` and `fixit`, but only do it after the class' __init__ has gone through it and prepared it.
        e = cls(error, fixit, **info)
        e.error = e.error.format(**e.info)
        if e.fixit:
            e.fixit = e.fixit.format(**e.info)
        return e

    @property
    def name(self):
        """ Name of the exception class """
        return self.__class__.__name__

    def headers(self, headers: dict):
        """ Additional headers to add to the HTTP response """
        self._response_headers = headers
        return self



""" Class-based views for FastAPI

Class-based views are implemented as @class_based_view() decorator.

Example:
    router = APIRouter()

    @class_based_view(
        router,
        Route(view_name='list', methods='GET', path='/'),  # or: ('list', 'GET', '/')
        Route(view_name='test', methods='GET', path='/test'),
    )
    class TestView:
        def __init__(self, current_user: models.User = Depends(current_user)):
            self.current_user = current_user

        def list(self):
            return {'users': []}

        def test(self, input: str):
            return {
                'current_user.id': self.current_user.id,
                'input': input,
            }

        @api_route.get('/self-test')
        def custom(self):
            return {'status': 'ok'}
"""
from __future__ import annotations

import dataclasses

import pydantic
from apiens.util import decomarker
from fastapi import APIRouter, FastAPI
from fastapi import Depends
from pydantic.utils import lenient_issubclass
from typing import List, Callable, Sequence, Union, Iterable, TypeVar, Optional, get_type_hints

from .patch_func import set_parameter_default, copy_func

ClsT = TypeVar('ClsT', bound=type)


class class_based_view:
    """ Decorator for classes to enable using their methods as class-based views """

    def __init__(self, router: Union[APIRouter, FastAPI], *routes: Union[Sequence, Route]):
        """ Decorate a class-based view

        Args:
            router: The APIRouter to register the routes on
            *routes: List of routes registerd on this class.
                Example: Route('list', '/', 'GET')
                Example: ('list', '/', 'GET')
                Alternatively, you can decorate individual methods using @api_route.get()
        """
        self.router = router
        self.routes: List[Route] = [
            # make Route() objects from tuples, if necessary
            route if isinstance(route, Route) else Route(*route)
            for route in routes
        ]

    def __call__(self, ViewCls: type):
        """ Decorator call """
        self.bind_to_class(ViewCls)

        # Transparently return the very same class.
        # The decorator is completely dissolved.
        return ViewCls

    def bind_to_class(self, ViewCls: type):
        """ Called when the class is known """
        # Discover new routes
        self._find_routes_on_class(ViewCls)

        # Add class information
        self._bind_routes_to_class(ViewCls)

        # Process the class
        self._prepare_class(ViewCls)

        # Patch every route endpoint
        self._patch_route_endpoints(ViewCls)

        # Register routes on the router
        self.register_routes_on(self.router, self.routes)

    @classmethod
    def register_routes_on(cls, router: APIRouter, routes: Iterable[Route]):
        """ Register all routes on the given router """
        for route in routes:
            if route.view_func_implemented:
                route.add_endpoint_to_router(router)

    def _prepare_class(self, ViewCls: type):
        """ Prepare the class before the routes are processed """
        # Run hooks
        if issubclass(ViewCls, ViewBase):
            ViewCls._customize_class_based_view(self)

    def _find_routes_on_class(self, ViewCls: type):
        """ Get @api_route()-decorated methods from the class and add them to self.routes  """
        self.routes.extend(
            decorated_endpoint.route
            for decorated_endpoint in api_route.all_decorated_from(ViewCls, inherited=True)
        )

    def _bind_routes_to_class(self, ViewCls: type):
        """ Add class information to every Route """
        for route in self.routes:
            route.bind_to_class(ViewCls)

    def _patch_route_endpoints(self, ViewCls: type):
        """ As soon as the class is known, patch every route to make sure it works with FastAPI """
        # Patch
        for route in self.routes:
            # Patch the endpoint
            if route.view_func_implemented:
                self._patch_endpoint(route)
                self._augment_endpoint_route(route)

    def _patch_endpoint(self, route: Route):
        """ Patch an endpoint to become compatible with FastAPI """
        route.view_func = copy_func(route.view_func)

        # By default, FastAPI will see the `self` argument and think it's a dependency.
        # It actually is: a dependency on an instantiated class.
        # Let's declare it as such: `self = Depends(ViewCls)`.
        route.view_func = patch_method_fastapi_compatible(route.ViewCls, route.view_func)

    def _augment_endpoint_route(self, route: Route):
        """ Get information from the endpoint and augment route """
        # Get type hints
        # Note that this evaluates forward references. We supply:
        # * the module as the global namespace
        # * the class as the local namespace
        # This lets you use class attributes as forward references. Yeehaw!
        # NOTE: type hints are evaluated on a per-view basis: thanks to copy_func(),
        # every class is evaluated in its own context! :)
        annotations = get_type_hints(
            route.view_func,
            # globalns=sys.modules[route.ViewCls.__module__].__dict__,  # use viewfunc.__globals__
            localns=vars(route.ViewCls)
        )

        # If there is anything useful in the return annotation, use it with FastAPI
        if lenient_issubclass(annotations.get('return'), pydantic.BaseModel):
            route.api_route_kwargs.setdefault('response_model', annotations['return'])


class api_route(decomarker):
    """ Decorate API methods on a class-based view

    NOTE: these routes will be added *after* other routes in the class-based view.
    TODO: override flag?

    Example:
        @class_based_view(router)
        class View:
            @api_route.get('/')
            def get(self, query: str):
                ...
    """

    def __init__(self, methods: Union[str, Iterable[str]], path: str, **api_route_kwargs):
        """ Describes a Route of a class-based view

        Args:
            methods: HTTP methods to use. Example: 'GET' or ['GET', 'POST']
            path: Endpoint path. Example: '/{id}'
            **api_route_kwargs: Additional options for APIRouter.add_api_route()
        """
        super().__init__()
        self.route = Route(..., methods, path, **api_route_kwargs)

    def __call__(self, handler_method: MethodT) -> MethodT:
        try:
            return super().__call__(handler_method)
        finally:
            self.route.view_name = self.func_name

    # region Convenience methods

    @classmethod
    def get(cls, path: str) -> api_route:
        return cls(path, 'GET')

    @classmethod
    def post(cls, path: str) -> api_route:
        return cls(path, 'POST')

    @classmethod
    def put(cls, path: str) -> api_route:
        return cls(path, 'PUT')

    @classmethod
    def patch(cls, path: str) -> api_route:
        return cls(path, 'PATCH')

    @classmethod
    def delete(cls, path: str) -> api_route:
        return cls(path, 'DELETE')

    # endregion


class ViewBase:
    """ Base for ClassBasedViews

    Note that @class_based_view can work with absolutely any class.
    However, if your class wants to customize the routes before they are registered, you have to inherit
    from this class and implement the hook methods.

    If you need something that is even more custom, then subclass @class_based_view
    """

    @classmethod
    def _customize_class_based_view(cls, decorator: class_based_view):
        """ Hook into @class_based_view and customize it before it is registered """
        raise NotImplementedError


@dataclasses.dataclass
class Route:
    """ Information about a route in a class-based view """

    def __init__(self, view_name: str, path: str, methods: Optional[Union[str, Iterable[str]]] = None, **api_route_kwargs):
        """ Describes a route of a class-based view

        Args:
            view_name: Name of the view function. E.g. 'get' will point to get()
            path: Endpoint path. Example: '/{id}'
            methods: HTTP methods to use. Example: 'GET' or ['GET', 'POST']. Default: 'GET'
            **api_route_kwargs: Additional options for APIRouter.add_api_route()
        """
        self.view_name = view_name
        self.path = path
        if methods is None:
            self.methods = None
        elif isinstance(methods, str):
            self.methods = [methods]
        else:
            self.methods = list(methods)
        self.api_route_kwargs = api_route_kwargs

    view_name: str
    path: str
    methods: List[str]
    api_route_kwargs: dict = dataclasses.field(default_factory=dict)

    def bind_to_class(self, ViewCls: type):
        """ Executed when the class is known """
        # Same: class, method
        self.ViewCls = ViewCls
        self.view_func = getattr(ViewCls, self.view_name)
        self.view_func_implemented = (self.view_func not in (None, NotImplemented))

    # Internal fields
    ViewCls: type = None
    view_func: Callable = None
    view_func_implemented: bool = None

    def add_endpoint_to_router(self, router: APIRouter):
        """ Add one route to the router """
        router.add_api_route(
            self.path,
            self.view_func,
            methods=self.methods,
            **self.api_route_kwargs
        )


class fastapi_compatible_method:
    """ A simple tool to implement a single method compatible with FastAPI

    It is a decorator that takes a method, and patches it to have the `self` argument
    to be declared as a dependency on the class constructor:

    * Arguments of the method become dependencies
    * Arguments of the __init__() method become dependencies as well
    * You can use the method as a dependency, or as an API endpoint.

    Example:

        class View:
            def __init__(self, arg: str):  # Arguments are dependencies (?arg)
                self.arg = arg

            @fastapi_compatible_method
            async def get(self, another_arg: str):  # arguments are dependencies (?another_arg)
                return {
                    'arg': self.arg,
                    'another_arg': another_arg,
                }

        app.get('/test')(View.get)
        # Check: http://localhost:8000/test?arg=value&another_arg=value

    It is a little awkward: you can't decorate the method with @app.get() directly: it's only possible
    to do it after the class has fully created.

    Thanks to: @dmontagu for the inspiration
    """

    # It is a descriptor: it wraps a method, and as soon as the method gets associated with a class,
    # it patches the `self` argument with the class dependency and dissolves without leaving a trace.

    def __init__(self, method: Callable):
        self.method = method

    def __set_name__(self, cls: type, method_name: str):
        # Patch the method and put it onto the class
        patched_method = patch_method_fastapi_compatible(cls, self.method)
        return setattr(cls, method_name, patched_method)


MethodT = TypeVar('MethodT', bound=Callable)


def patch_method_fastapi_compatible(Cls: type, method: MethodT) -> MethodT:
    """ Patch the method function to become compatible with FastAPI

    We only have to declare `self` as a dependency on the class itself: `self = Depends(cls)`.
    Using this hack, FastAPI provides the `self` argument as a dependency by instantiating a class.

    Args:
        Cls: the class the method is bound to
        method: the method function
    Returns:
        the same function, patched.
        If you want it clean, make a copy()
    """
    # By default, FastAPI will see the `self` argument and think it's a dependency.
    # It actually is: a dependency on an instantiated class.
    # Let's declare it as such: `self = Depends(ViewCls)`.
    patched_method = set_parameter_default(method, 'self', Depends(Cls))
    return patched_method




from __future__ import annotations

import functools
import inspect
from typing import Hashable, Callable, Union, ClassVar, Optional, Any

from apiens.util import decomarker
from .errors import BaseApplicationError
from .signature import Signature
from .util.documented_function import DocumentedFunction


class operation(decomarker):
    """ Decorator for functions that implement business-logic

    Example:

        from apiens import operation, di

        @operation()
        @di.kwargs(current_user=authenticated_user)
        def who_am_i(current_user: User):
            return current_user.login

    Example:

        @operation(id='user')
        @di.kwargs(ssn=Session)
        class UserLogic:
            def __init__(self, ssn: Session):
                self.ssn = ssn

            @operation()
            @di.signature()
            def list(self):
                return ssn.query(...)

            @operation()
            @di.signature()
            def create(self, user: schemas.User):
                ...

            @operation()
            @di.signature()
            def delete(self, id: int):
                ...
    """

    # A unique name of this operation.
    # It will be used to call it as a function
    operation_id: Union[str, Hashable]

    # Operation function's signature: argument types, defaults, etc
    signature: Signature

    # Operation's documentation
    doc: DocumentedFunction

    # Extra, custom, information about the operation
    info: dict

    def __init__(self,
                 operation_id: Optional[Union[str, Hashable]] = None,
                 **info):
        """

        Args:
            id: Operation id. Can be a string name, or a tuple of multiple names to mimic a tree-like structure.
                When `None`, is taken from the function's name3
            **info: Additional information to associate with this operation. Arbitrary.
        """
        super().__init__()
        self.operation_id = operation_id
        self.doc = DocumentedFunction(None, {}, None, {}, None)

        # Parse the docstring as soon as it becomes available
        self._parse_docstring = True

        # Custom extra info
        self.info: dict = info

    def decorator(self, func: Union[Callable, type]):
        # `id` defaults to function name
        if self.operation_id is None:
            self.operation_id = func.__name__

        # Read function signature into information
        self.signature = Signature(func)

        # Parse the function's docstring
        if self._parse_docstring:
            try:
                self.doc.parse_from_function_docstring(func)
            except Exception as e:
                raise ValueError(f'Error while parsing docstring of {func}') from e

        # If we've decorated a class, go through every method and tell it about it
        if inspect.isclass(func):
            # Remove the "self" argument from the constructor signature
            remove_first_key(self.signature.arguments)

            # Remove the "self" argument from every decorated method as well
            for method_operation in self.all_decorated_from(func, inherited=True):
                if not isinstance(func, staticmethod):
                    remove_first_key(method_operation.signature.arguments)

        # Done
        return super().decorator(func)

    def pluck_kwargs_from(self, kwargs: dict):
        """ Given a dict of many parameters, choose the ones that this operation needs

        This method is used with class-based views, where the input contains parameters for both
        the __init__() method of the class and the operation method.
        """
        return {
            name: kwargs[name]
            for name in self.signature.arguments
            if name in kwargs  # because some defaults might not be provided
        }

    def check_operation_documentation(self, *, fully_documented: bool = True):
        """ Check that all documentations make sense """
        # Check that every documented parameter is actually a known parameter
        documented_parameter_names = set(self.doc.parameters)
        function_parameter_names = set(inspect.signature(self.func).parameters)
        mistaken_parameter_names = documented_parameter_names - function_parameter_names

        # Error message
        s = 's' if len(mistaken_parameter_names) > 1 else ''
        assert not mistaken_parameter_names, (
            f'Unknown parameter name{s} documented for the function: {mistaken_parameter_names}. '
            f'This is probably a typo. Please document this function properly: {self.func}.'
        )

        # Fully documented?
        # TODO: disable "fully_documented" on a per-method basis
        # TODO: inherit documentation (errors, parameters, function doc) from the class and from the implementation on the parent
        if fully_documented:
            self._assert_is_fully_documented()

            # Also, wrap the function to check what sort of errors does it throw
            self.func = self.wrap_func_check_thrown_errors_are_documented(self.func)

    def _assert_is_fully_documented(self):
        """ Check that the function is fully documented. Used in projects by documenting freaks.

        It will test that:

        * The function is documented
        * Every parameter and the return value are both annotated and documented

        Args:
            as_class_method: Validate it as a class method (i.e. ignoring the first parameter)

        Raises:
            AssertionError if anything failed.
        """
        errors = []

        # Class? Drop a few checks.
        is_class = inspect.isclass(self.func)

        # Check that the function itself is documented
        if not self.doc.function:
            errors.append(f"The function is not documented. Please add a docstring.")

        # Check that its return value is documented
        if not is_class and not self.doc.result:
            errors.append(f"The return value is not documented. Please add the 'Returns' section.")

        # Get its signature
        signature = operation.get_from(self.func).signature

        # Check that its return value is typed
        if not is_class and signature.return_type is Any:
            errors.append(f"The return type is unknown. Please provide a return type annotation.")

        # Check every parameter's type and documentation
        for name, type_ in signature.arguments.items():
            if type_ is Any:
                errors.append(f"Parameter {name!r} type is unknown. "
                              f"Please put a type annotation on it.")
            if name not in self.doc.parameters:
                errors.append(f"Parameter {name!r} is not documented. "
                              f"Please add an 'Args' section for it")

        # We can't validate errors.
        # But error_validator_decorator() can make a wrapper that will validate them at runtime :)

        # Report errors
        if errors:
            errors_str = ''.join(f'\t* {error}\n' for error in errors)
            raise AssertionError(
                f'Function is not fully documented: {self.func}\n'
                f'Errors: \n'
                f'{errors_str}'
            )

    def wrap_func_check_thrown_errors_are_documented(self, func: Callable) -> Callable:
        """ A wrapper that will make sure that every error raised by the function is documented """
        # Wrap the function
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Call the function
            try:
                return func(*args, **kwargs)
            # Catch errors, see if they're documented
            except BaseApplicationError as e:
                if type(e) not in self.doc.errors:
                    raise AssertionError(
                        f"Function {func} raised an undocumented error: {e.name!r}. "
                        f"Please document the error in a `Raises:` docstring section"
                    ) from e
                raise

        # Done
        return wrapper


def remove_first_key(d: dict):
    """ Remove the first key from a dictionary. Used to remove `self` """
    first_key = list(d)[0]
    d.pop(first_key)



import inspect
import typing
from typing import Callable, Mapping, Union, Any, FrozenSet, Literal

from apiens import di
from dataclasses import dataclass


@dataclass
class Signature:
    """ A function's type signature: the information about its input and output types

    This signature information can be used to generate validation schemas, e.g. attrs or pydantic.
    """

    # Arguments that will be provided by the DI Injector
    # Mapping: { name => type }
    provided_arguments: Mapping[str, Union[type, Literal[Any]]]

    # Arguments that must be provided by the caller
    # Mapping: { name => type }
    arguments: Mapping[str, Union[type, Literal[Any]]]

    # Argument and their default values
    # Mapping: { name => default value }
    argument_defaults: Mapping[str, Any]

    # Return value type, as read from the function's signature
    # Value: type
    return_type: Union[type, Literal[Any]]

    def __init__(self, func: Callable):
        # Get the names of dependencies that will be provided
        provided_names = get_provided_names(func)

        # If it's a class, go to its constructor instead.
        # We only do this after `get_provided_names()` to make sure
        # we support both decorated classes and decorated constructors
        if inspect.isclass(func):
            func = func.__init__
            provided_names |= get_provided_names(func)

        # Get all annotations.
        # Note that un-annotated parameters will be missing.
        parameter_annotations = typing.get_type_hints(func)
        self.return_type = parameter_annotations.pop('return', Any)

        # Prepare
        self.provided_arguments: dict = {}
        self.arguments: dict = {}
        self.argument_defaults: dict = {}

        # Read every parameter from the function
        for name, parameter in inspect.signature(func).parameters.items():
            # Skip variadic arguments
            if parameter.kind in (parameter.VAR_KEYWORD, parameter.VAR_POSITIONAL):
                continue

            # Get the type. Default to `Any`
            type_ = parameter_annotations.get(parameter.name, Any)

            # Where does this argument go? `provided_arguments` or `arguments`?
            # An argument from the DI Injector
            if name in provided_names:
                self.provided_arguments[name] = type_
            # A user-provided argument
            else:
                self.arguments[name] = type_

                # Defaults
                if parameter.default is not parameter.empty:
                    self.argument_defaults[name] = parameter.default


def get_provided_names(func: Callable) -> FrozenSet[str]:
    """ Get names of the arguments that will be provided by the DI Injector """
    di_marker = di.resolvable_marker.get_from(func)
    if di_marker:
        return di_marker.resolvable.provided_names()
    else:
        return frozenset()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apiens',
 'apiens.di',
 'apiens.testing',
 'apiens.tools',
 'apiens.tools.fastapi',
 'apiens.tools.fastapi.class_based_view',
 'apiens.util',
 'apiens.via',
 'apiens.via.fastapi',
 'apiens.views',
 'apiens.views.crud',
 'apiens.views.mongoquery_crud']

package_data = \
{'': ['*']}

extras_require = \
{'crud': ['pydantic>=1.6.1,<2.0.0',
          'sqlalchemy>=1.3.19,<2.0.0',
          'mongosql>=2.0.11,<3.0.0',
          'sa2schema>=0.1.3,<0.2.0',
          'blinker>=1.4,<2.0'],
 'fastapi': ['fastapi>=0.61.1,<0.62.0']}

setup_kwargs = {
    'name': 'apiens',
    'version': '1.0.1',
    'description': 'Business-logic layer for your API',
    'long_description': None,
    'author': 'Mark Vartanyan',
    'author_email': 'kolypto@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kolypto/py-apiens',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from enum import Enum

from typing import Type, Any


class TitledEnum(Enum):
    """ An Enum with both a value and a title """

    def __new__(cls, value: Any, title: str):
        v = object.__new__(cls)
        v._value_ = value
        v.title = title
        return v


def titled(title, *, description: str = ''):
    """ Give enumeration a title and a description

    Note that this can be used independently of `TitledEnum`

    Example:
        @titled(_('Message direction'))
        class Direction(TitledEnum):
            IN = -1, _('Incoming')
            OUT = +1, _('Outgoing')
    """
    def EnumClassWrapper(Enum: Type[Enum]):
        Enum.__title__ = title
        Enum.__description__ = description
        return Enum
    return EnumClassWrapper


def get_title(Enum: Type[Enum]) -> str:
    """ Get the title of a titled Enum """
    return getattr(Enum, '__title__', '(not set)')


def get_description(Enum: Type[Enum]) -> str:
    """ Get the description of a titled Enum """
    return getattr(Enum, '__description__', '')
